# optionsbar_filters.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
from .abstract_optionsbar import AbstractOptionsBar
from .utilities_units import utilities_add_unit_to_spinbtn

class OptionsBarFilters(AbstractOptionsBar):
	__gtype_name__ = 'OptionsBarFilters'

	def __init__(self, window, filters_tool):
		super().__init__()
		self.window = window
		self.filters_tool = filters_tool

		self._build_ui('transform/abstract-optionsbar-transform.ui')
		builder = self._hydrate_transform_tool('transform/optionsbar-filters.ui')

		self.menu_btn = builder.get_object('menu_btn')
		self._togglable_btn = self.menu_btn
		self.menu_btn.set_menu_model(self.filters_tool.get_options_model())
		self._menu_label = builder.get_object('menu_label')
		self._menu_icon = builder.get_object('menu_icon')

	def add_spinbtn(self, caption, adj_as_array, spin_chars, unit):
		widget_label = Gtk.Label(label=caption)
		widget_spinbtn = Gtk.SpinButton(tooltip_text=caption)
		adj = Gtk.Adjustment()
		# it's [value, lower, upper, step_increment, page_increment, page_size]
		adj.configure(*adj_as_array)
		widget_spinbtn.set_adjustment(adj)
		utilities_add_unit_to_spinbtn(widget_spinbtn, spin_chars, unit)

		self.centered_box.add(widget_label)
		self.centered_box.add(widget_spinbtn)
		return widget_label, widget_spinbtn

	############################################################################

	def init_adaptability(self):
		super().init_adaptability()
		self._menu_icon.set_visible(False)
		widgets_size = self.filters_tool.get_max_filter_width()
		temp_limit_size = self.menu_btn.get_preferred_width()[0] + \
		                  50 + widgets_size + \
		                  self.cancel_btn.get_preferred_width()[0] + \
		                  self.help_btn.get_preferred_width()[0] + \
		                  self.apply_btn.get_preferred_width()[0]
		self._set_limit_size(temp_limit_size)

	def on_filter_changed(self):
		self.set_compact(self._is_narrow)
		self.window.set_window_subtitles()
		# self.menu_label.set_label(self.filters_tool.type_label) # XXX width???

	def set_compact(self, state):
		super().set_compact(state)
		self._menu_label.set_visible(not state)
		self._menu_icon.set_visible(state)
		self.filters_tool.set_filters_compact(state)

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

from .abstract_optionsbar import AbstractOptionsBar

class OptionsBarSelection(AbstractOptionsBar):
	__gtype_name__ = 'OptionsBarSelection'

	def __init__(self, window):
		super().__init__()
		self.window = window
		builder = self._build_ui('selection/optionsbar-selection.ui')

		self.import_box = builder.get_object('import_box')
		self.clipboard_box = builder.get_object('clipboard_box')
		self.actions_btn = builder.get_object('actions_btn')
		self.actions_btn_long = builder.get_object('actions_btn_long')
		self._togglable_btn = self.actions_btn

		self.options_long_box = builder.get_object('options_long_box')
		self.options_short_box = builder.get_object('options_short_box')

		self.minimap_btn = builder.get_object('minimap_btn')
		self.minimap_label = builder.get_object('minimap_label')
		self.minimap_arrow = builder.get_object('minimap_arrow')

	def get_minimap_btn(self):
		return self.minimap_btn

	def set_minimap_label(self, label):
		self.minimap_label.set_label(label)

	def middle_click_action(self):
		self.window.lookup_action('new_tab_selection').activate()

	############################################################################

	def init_adaptability(self):
		super().init_adaptability()
		temp_limit_size = self.import_box.get_preferred_width()[0] + \
		                    self.clipboard_box.get_preferred_width()[0] + \
		                      self.actions_btn.get_preferred_width()[0] + \
		                 self.options_long_box.get_preferred_width()[0] + \
		                         self.help_btn.get_preferred_width()[0] + \
		                      self.minimap_btn.get_preferred_width()[0]
		self._set_limit_size(temp_limit_size)

	def set_compact(self, state):
		super().set_compact(state)
		self.actions_btn_long.set_visible(state)
		self.import_box.set_visible(not state)
		self.actions_btn.set_visible(not state)

		self.options_long_box.set_visible(not state)
		self.options_short_box.set_visible(state)
		if state:
			self._togglable_btn = self.actions_btn_long
		else:
			self._togglable_btn = self.actions_btn

		self.minimap_arrow.set_visible(not state)

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo, threading
# from datetime import datetime # Not actually needed, just to measure perfs

class BlurType(int):
	INVALID = -1
	AUTO = 0
	PX_BOX = 1
	PX_BOX_MULTI = 2
	CAIRO_REPAINTS = 3
	TILES = 4

class BlurDirection(int):
	INVALID = -1
	BOTH = 0
	HORIZONTAL = 1
	VERTICAL = 2

################################################################################

def utilities_blur_surface(surface, radius, blur_type, blur_direction):
	"""This is the 'official' method to access the blur algorithms.
	The third argument is an integer corresponding to the BlurType enumeration.
	The 4th one is an integer corresponding to the BlurDirection enumeration."""
	radius = int(radius)
	if radius < 1:
		return surface
	blurred_surface = None
	# time0 = datetime.now()
	# print('blurring begins, using algo ', blur_type, '-', blur_direction)

	if blur_type == BlurType.INVALID:
		return surface
	elif blur_type == BlurType.AUTO:
		blur_type = BlurType.PX_BOX
		# XXX c'est nul ça mdr, mais bon c'est peu utilisé

	if blur_type == BlurType.PX_BOX:
		blurred_surface = _generic_px_box_blur(surface, radius, blur_direction)
	elif blur_type == BlurType.PX_BOX_MULTI:
		blurred_surface = _generic_multi_threaded_blur(surface, radius, blur_direction)
	elif blur_type == BlurType.CAIRO_REPAINTS:
		blurred_surface = _generic_cairo_blur(surface, radius, blur_direction)
	elif blur_type == BlurType.TILES:
		blurred_surface = _generic_tiled_blur(surface, radius, blur_direction)

	# time1 = datetime.now()
	# print('blurring ended, total time:', time1 - time0)
	return blurred_surface

################################################################################
# BlurType.PX_BOX ##############################################################

def _generic_px_box_blur(surface, radius, blur_direction):
	w = surface.get_width()
	h = surface.get_height()
	channels = 4 # ARGB
	if radius > w - 1 or radius > h - 1:
		return surface

	# this code a modified version of this https://github.com/elementary/granite/blob/14e3aaa216b61f7e63762214c0b36ee97fa7c52b/lib/Drawing/BufferSurface.vala#L230
	# the main differences (aside of the language) is the poor attempt to use
	# multithreading (i'm quite sure the access to buffers are not safe at all).
	# The 2 phases of the algo have been separated to allow directional blur.
	original = cairo.ImageSurface(cairo.Format.ARGB32, w, h)
	cairo_context = cairo.Context(original)
	# cairo_context.set_operator(cairo.Operator.SOURCE)
	cairo_context.set_source_surface(surface, 0, 0)
	cairo_context.paint() # XXX is this copy useful?
	original.flush()
	pixels = original.get_data()

	buffer0 = [None] * (w * h * channels)
	vmin = [None] * max(w, h)
	vmax = [None] * max(w, h)
	div = 2 * radius + 1
	dv = [None] * (256 * div)
	for i in range(0, len(dv)):
		dv[i] = int(i / div)

	iterations = 1
	while iterations > 0:
		iterations = iterations - 1
		if blur_direction == BlurDirection.VERTICAL:
			for i in range(0, len(pixels)):
				buffer0[i] = pixels[i]
		else:
			_box_blur_1st_phase(w, h, channels, radius, pixels, buffer0, vmin, vmax, dv)
		# print('end of the 1st phase…', datetime.now() - time0)
		if blur_direction == BlurDirection.HORIZONTAL:
			for i in range(0, len(buffer0)):
				pixels[i] = buffer0[i]
		else:
			_box_blur_2nd_phase(w, h, channels, radius, pixels, buffer0, vmin, vmax, dv)
	return original

# this code a modified version of a naïve approach to box blur, copied from
# here: https://github.com/elementary/granite/blob/14e3aaa216b61f7e63762214c0b36ee97fa7c52b/lib/Drawing/BufferSurface.vala#L230
# The 2 phases of the algo have been separated to allow directional blur.

def _box_blur_1st_phase(w, h, channels, radius, pixels, buff0, vmin, vmax, dv):
	"""Horizontal blurring"""
	for x in range(0, w):
		vmin[x] = min(x + radius + 1, w - 1)
		vmax[x] = max(x - radius, 0)
	for y in range(0, h):
		cur_pixel = y * w * channels
		asum = radius * pixels[cur_pixel + 0]
		rsum = radius * pixels[cur_pixel + 1]
		gsum = radius * pixels[cur_pixel + 2]
		bsum = radius * pixels[cur_pixel + 3]
		for i in range(0, radius+1):
			asum += pixels[cur_pixel + 0]
			rsum += pixels[cur_pixel + 1]
			gsum += pixels[cur_pixel + 2]
			bsum += pixels[cur_pixel + 3]
			cur_pixel += channels
		cur_pixel = y * w * channels
		for x in range(0, w):
			p1 = (y * w + vmin[x]) * channels
			p2 = (y * w + vmax[x]) * channels
			buff0[cur_pixel + 0] = dv[asum]
			buff0[cur_pixel + 1] = dv[rsum]
			buff0[cur_pixel + 2] = dv[gsum]
			buff0[cur_pixel + 3] = dv[bsum]
			asum += pixels[p1 + 0] - pixels[p2 + 0]
			rsum += pixels[p1 + 1] - pixels[p2 + 1]
			gsum += pixels[p1 + 2] - pixels[p2 + 2]
			bsum += pixels[p1 + 3] - pixels[p2 + 3]
			cur_pixel += channels

def _box_blur_2nd_phase(w, h, channels, radius, pixels, buff0, vmin, vmax, dv):
	"""Vertical blurring"""
	for y in range(0, h):
		vmin[y] = min(y + radius + 1, h - 1) * w
		vmax[y] = max (y - radius, 0) * w
	for x in range(0, w):
		cur_pixel = x * channels
		asum = radius * buff0[cur_pixel + 0]
		rsum = radius * buff0[cur_pixel + 1]
		gsum = radius * buff0[cur_pixel + 2]
		bsum = radius * buff0[cur_pixel + 3]
		for i in range(0, radius+1):
			asum += buff0[cur_pixel + 0]
			rsum += buff0[cur_pixel + 1]
			gsum += buff0[cur_pixel + 2]
			bsum += buff0[cur_pixel + 3]
			cur_pixel += w * channels
		cur_pixel = x * channels
		for y in range(0, h):
			p1 = (x + vmin[y]) * channels
			p2 = (x + vmax[y]) * channels
			pixels[cur_pixel + 0] = dv[asum]
			pixels[cur_pixel + 1] = dv[rsum]
			pixels[cur_pixel + 2] = dv[gsum]
			pixels[cur_pixel + 3] = dv[bsum]
			asum += buff0[p1 + 0] - buff0[p2 + 0]
			rsum += buff0[p1 + 1] - buff0[p2 + 1]
			gsum += buff0[p1 + 2] - buff0[p2 + 2]
			bsum += buff0[p1 + 3] - buff0[p2 + 3]
			cur_pixel += w * channels

################################################################################
# BlurType.PX_BOX_MULTI ########################################################

def _generic_multi_threaded_blur(surface, radius, blur_direction):
	"""Experimental multi-threaded blur. The parameter `blur_direction` will be
	ignored (it always blurs in both directions)."""
	w = surface.get_width()
	h = surface.get_height()
	channels = 4 # ARGB
	if radius > w - 1 or radius > h - 1:
		return

	original = cairo.ImageSurface(cairo.Format.ARGB32, w, h)
	cairo_context = cairo.Context(original)
	# cairo_context.set_operator(cairo.Operator.SOURCE)
	cairo_context.set_source_surface(surface, 0, 0)
	cairo_context.paint()
	original.flush()
	pixels = original.get_data()

	buffer0 = [None] * (w * h * channels)
	vmin = [None] * max(w, h)
	vmax = [None] * max(w, h)
	div = 2 * radius + 1
	dv = [None] * (256 * div)
	for i in range(0, len(dv)):
		dv[i] = int(i / div)

	full_buff = _box_blur_1st_phase_multi(w, h, channels, radius, pixels, vmin, vmax, dv)
	# print('end of the 1st phase…', datetime.now() - time0)
	_box_blur_2nd_phase(w, h, channels, radius, pixels, full_buff, vmin, vmax, dv)

	return original

# this code a modified version of a naïve approach to box blur, copied from
# here: https://github.com/elementary/granite/blob/14e3aaa216b61f7e63762214c0b36ee97fa7c52b/lib/Drawing/BufferSurface.vala#L230
# the main differences (aside of the language) is the poor attempt to use
# multithreading (i'm quite sure the access to buffers are not safe at all)
# during the first phase. Multithreading of the second phase has not even been
# tried because this multithreaded version is slower than _box_blur_1st_phase

def _box_blur_1st_phase_multi(w, h, channels, radius, pixels, vmin, vmax, dv):
	NB_THREADS = 4
	t1 = [None] * NB_THREADS
	full_buffer = []
	buffers = []
	y_end = 0
	for x in range(0, w):
		vmin[x] = min(x + radius + 1, w - 1)
		vmax[x] = max(x - radius, 0)
	for t in range(0, NB_THREADS):
		y_start = y_end
		if t == NB_THREADS - 1:
			y_end = h
		else:
			y_end = y_start + int(h / NB_THREADS)
		buff_size = w * (y_end - y_start) * channels
		buff = [0] * buff_size
		buffers.append(buff)
		t1[t] = threading.Thread(target=_blur_rows3, args=(x, y_start, y_end, \
		                     w, channels, radius, pixels, buff, vmin, vmax, dv))
	for t in range(0, NB_THREADS):
		t1[t].start()
	for t in range(0, NB_THREADS):
		t1[t].join()
		full_buffer += buffers[t]
	return full_buffer

def _blur_rows3(x, y0, y1, w, channels, radius, pixels, buff0, vmin, vmax, dv):
	# print('row thread with', y0, y1, '(begin)')
	diff = y0 * w * channels
	for y in range(y0, y1):
		cur_pixel = y * w * channels
		asum = radius * pixels[cur_pixel + 0]
		rsum = radius * pixels[cur_pixel + 1]
		gsum = radius * pixels[cur_pixel + 2]
		bsum = radius * pixels[cur_pixel + 3]
		for i in range(0, radius+1):
			asum += pixels[cur_pixel + 0]
			rsum += pixels[cur_pixel + 1]
			gsum += pixels[cur_pixel + 2]
			bsum += pixels[cur_pixel + 3]
			cur_pixel += channels
		cur_pixel = y * w * channels - diff
		for x in range(0, w):
			p1 = (y * w + vmin[x]) * channels
			p2 = (y * w + vmax[x]) * channels
			buff0[cur_pixel + 0] = dv[asum]
			buff0[cur_pixel + 1] = dv[rsum]
			buff0[cur_pixel + 2] = dv[gsum]
			buff0[cur_pixel + 3] = dv[bsum]
			asum += pixels[p1 + 0] - pixels[p2 + 0]
			rsum += pixels[p1 + 1] - pixels[p2 + 1]
			gsum += pixels[p1 + 2] - pixels[p2 + 2]
			bsum += pixels[p1 + 3] - pixels[p2 + 3]
			cur_pixel += channels
	# print('row thread with', y0, y1, '(end)')

################################################################################
# BlurType.CAIRO_REPAINTS ######################################################

def _generic_cairo_blur(surface, radius, blur_direction):
	if blur_direction == BlurDirection.HORIZONTAL:
		surface = _cairo_directional_blur(surface, radius, False)
	elif blur_direction == BlurDirection.VERTICAL:
		surface = _cairo_directional_blur(surface, radius, True)
	else:
		# XXX with some big radius, it's visible that it's just a sequence of
		# 2 directional blurs instead of an actual algorithm
		surface = _cairo_directional_blur(surface, radius, True)
		surface = _cairo_directional_blur(surface, radius, False)
	return surface

# Weird attempt to produce a blurred image using cairo. I mean ok, the image is
# blurred, and with amazing performances, but the quality is not convincing and
# the result when the area has (semi-)transparency really sucks.

def _cairo_directional_blur(surface, radius, is_vertical):
	cairo_context = cairo.Context(surface)
	if radius < 10:
		step = 1
		alpha = min(0.9, step / radius)
	elif radius < 15:
		step = 1
		alpha = min(0.9, (0.5 + step) / radius)
	else:
		step = int(radius / 6) # why 6? mystery
		# cette optimisation donne de légers glitchs aux grands radius, qui ne
		# sont de toutes manières pas beaux car on voit en partie à travers
		alpha = min(0.9, (1 + step) / radius)
	for i in range(-1 * radius, radius, step):
		if is_vertical:
			cairo_context.set_source_surface(surface, 0, i)
		else:
			cairo_context.set_source_surface(surface, i, 0)
		cairo_context.paint_with_alpha(alpha)
	surface.flush()
	return surface

################################################################################
# BlurType.TILES ###############################################################

def _generic_tiled_blur(surface, radius, blur_direction):
	if blur_direction == BlurDirection.HORIZONTAL:
		tile_width = radius
		tile_height = 1
	elif blur_direction == BlurDirection.VERTICAL:
		tile_width = 1
		tile_height = radius
	else:
		tile_width = radius
		tile_height = radius
	return _get_tiled_surface(surface, tile_width, tile_height)

def _get_tiled_surface(surface, tile_width, tile_height):
	w = surface.get_width()
	h = surface.get_height()
	channels = 4 # ARGB
	pixels = surface.get_data()
	pixel_max = w * h * channels

	for x in range(0, w, tile_width):
		for y in range(0, h, tile_height):
			current_pixel = (x + (w * y)) * channels
			if current_pixel >= pixel_max:
				continue
			tile_b = pixels[current_pixel + 0]
			tile_g = pixels[current_pixel + 1]
			tile_r = pixels[current_pixel + 2]
			tile_a = pixels[current_pixel + 3]
			for tx in range(0, tile_width):
				for ty in range(0, tile_height):
					current_pixel = ((x + tx) + (w * (y + ty))) * channels
					if current_pixel >= pixel_max:
						continue
					if current_pixel >= (w * (y + ty + 1)) * channels:
						# the current tile is out of the surface, this guard
						# clause avoids corrupting the next tile
						continue
					pixels[current_pixel + 0] = tile_b
					pixels[current_pixel + 1] = tile_g
					pixels[current_pixel + 2] = tile_r
					pixels[current_pixel + 3] = tile_a
			# end of one tile
	# end of the "for each tile"
	return surface

################################################################################




# tool_arc.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from .abstract_classic_tool import AbstractClassicTool
from .utilities_paths import utilities_add_arrow_triangle

class ToolArc(AbstractClassicTool):
	__gtype_name__ = 'ToolArc'

	def __init__(self, window, **kwargs):
		super().__init__('arc', _("Curve"), 'tool-arc-symbolic', window)
		self.use_operator = True

		# Default values
		self._cap_id = cairo.LineCap.ROUND
		self._1st_segment = None
		self._dashes_type = 'none'
		self._arrow_type = 'none'
		self._use_outline = False

		self.add_tool_action_enum('line_shape', 'round')
		self.add_tool_action_enum('dashes-type', self._dashes_type)
		self.add_tool_action_enum('arrow-type', self._arrow_type)
		self.add_tool_action_boolean('pencil-outline', self._use_outline)

	def give_back_control(self, preserve_selection):
		self._1st_segment = None
		self.restore_pixbuf()

	############################################################################
	# Options ##################################################################

	def set_active_shape(self):
		state_as_string = self.get_option_value('line_shape')
		if state_as_string == 'thin':
			self._cap_id = cairo.LineCap.BUTT
		else:
			self._cap_id = cairo.LineCap.ROUND

	def get_options_label(self):
		return _("Curve options")

	def get_editing_tips(self):
		self._use_outline = self.get_option_value('pencil-outline')
		self._dashes_type = self.get_option_value('dashes-type')
		self._arrow_type = self.get_option_value('arrow-type')
		self.set_active_shape()
		is_arrow = self._arrow_type != 'none'
		use_dashes = self._dashes_type != 'none'

		label_segments = self.label + " - "
		if self._1st_segment is not None:
			label_segments += _("Draw a second segment to complete the curve")
		else:
			label_segments = None

		label_options = self.label + " - "
		if is_arrow and use_dashes:
			label_options += _("Dashed arrow")
		elif is_arrow:
			label_options += _("Arrow")
		elif use_dashes:
			label_options += _("Dashed")
		else:
			label_options = None

		if self.get_image().get_mouse_is_pressed():
			label_modifier_alt = None
		else:
			label_modifier_alt = self.label + " - " + \
			                     _("Press <Alt> to toggle the 'outline' option")

		full_list = [label_segments, label_options, label_modifier_alt]
		return list(filter(None, full_list))

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.set_common_values(event.button, event_x, event_y)

		self.update_modifier_state(event.state)
		if 'ALT' in self._modifier_keys:
			self._use_outline = not self._use_outline

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		cairo_context = self.get_context()
		if self._1st_segment is None:
			cairo_context.move_to(self.x_press, self.y_press)
			cairo_context.line_to(event_x, event_y)
		else:
			cairo_context.move_to(self._1st_segment[0], self._1st_segment[1])
			cairo_context.curve_to(self._1st_segment[2], self._1st_segment[3], \
			                       self.x_press, self.y_press, event_x, event_y)
		self._path = cairo_context.copy_path()

		if not render:
			return
		operation = self.build_operation(event_x, event_y)
		self.do_tool_operation(operation)

	def on_release_on_area(self, event, surface, event_x, event_y):
		if self._1st_segment is None:
			self._1st_segment = (self.x_press, self.y_press, event_x, event_y)
			return
		else:
			cairo_context = self.get_context()
			cairo_context.move_to(self._1st_segment[0], self._1st_segment[1])
			cairo_context.curve_to(self._1st_segment[2], self._1st_segment[3], \
			                       self.x_press, self.y_press, event_x, event_y)
			self._1st_segment = None

		self._path = cairo_context.copy_path()
		operation = self.build_operation(event_x, event_y)
		self.apply_operation(operation)

	############################################################################

	def build_operation(self, event_x, event_y):
		operation = {
			'tool_id': self.id,
			'rgba': self.main_color,
			'rgba2': self.secondary_color,
			'antialias': self._use_antialias,
			'operator': self._operator,
			'line_width': self.tool_width,
			'line_cap': self._cap_id,
			'dashes': self._dashes_type,
			'arrow': self._arrow_type,
			'outline': self._use_outline,
			'path': self._path,
			'x_release': event_x,
			'y_release': event_y,
			'x_press': self.x_press,
			'y_press': self.y_press
		}
		return operation

	def do_tool_operation(self, operation):
		cairo_context = self.start_tool_operation(operation)

		cairo_context.set_operator(operation['operator'])
		line_width = operation['line_width']
		cairo_context.set_line_width(line_width)
		self.set_dashes_and_cap(cairo_context, line_width, \
		                             operation['dashes'], operation['line_cap'])

		if operation['arrow'] == 'double':
			for pts in operation['path']:
				# how to do without a for???
				if(pts[0] == cairo.PathDataType.MOVE_TO):
					x1 = pts[1][0]
					y1 = pts[1][1]
				else:
					x2 = pts[1][0]
					y2 = pts[1][1]
			utilities_add_arrow_triangle(cairo_context, x1, y1, x2, y2, line_width)

		cairo_context.append_path(operation['path'])

		if operation['arrow'] != 'none':
			x1 = operation['x_press']
			y1 = operation['y_press']
			x2 = operation['x_release']
			y2 = operation['y_release']
			utilities_add_arrow_triangle(cairo_context, x2, y2, x1, y1, line_width)

		if operation['outline']:
			cairo_context.set_source_rgba(*operation['rgba2'])
			cairo_context.set_line_width(line_width * 1.2 + 2)
			cairo_context.stroke_preserve()

		cairo_context.set_line_width(line_width)
		cairo_context.set_source_rgba(*operation['rgba'])
		cairo_context.stroke()

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

from gi.repository import Gtk

################################################################################

def utilities_add_unit_to_spinbtn(spinbutton, width_chars, unit):
	spinbutton.set_width_chars(width_chars + 3)
	if unit == 'px':
		# To translators: it's a measure unit, it appears in tooltips over
		# numerical inputs
		_add_spinbutton_icon(spinbutton, 'unit-pixels-symbolic', _("pixels"))
	elif unit == '%':
		# To translators: it appears in tooltips over numerical inputs
		_add_spinbutton_icon(spinbutton, 'unit-percents-symbolic', _("percents"))
	elif unit == '°':
		# To translators: it's the angle measure unit, it appears in a tooltip
		# over a numerical input
		_add_spinbutton_icon(spinbutton, 'unit-degrees-symbolic', _("degrees"))

def _add_spinbutton_icon(spinbutton, icon, tooltip):
	p = Gtk.EntryIconPosition.SECONDARY
	spinbutton.set_icon_from_icon_name(p, icon)
	spinbutton.set_icon_tooltip_text(p, tooltip)
	spinbutton.set_icon_sensitive(p, False)

################################################################################




# tool_line.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo, math
from .abstract_classic_tool import AbstractClassicTool
from .utilities_paths import utilities_add_arrow_triangle

class ToolLine(AbstractClassicTool):
	__gtype_name__ = 'ToolLine'

	def __init__(self, window, **kwargs):
		super().__init__('line', _("Line"), 'tool-line-symbolic', window)
		self.use_operator = True

		self._use_outline = False
		self._dashes_type = 'none'
		self._arrow_type = 'none'
		self._use_gradient = False
		# Lock the tool to only draw orthogonal lines (multiples of 45°)
		self._ortholock = False # Unrelated to Maïté's ortolan.

		self.add_tool_action_enum('line_shape', 'round')
		self.add_tool_action_enum('dashes-type', self._dashes_type)
		self.add_tool_action_enum('arrow-type', self._arrow_type)
		self.add_tool_action_boolean('use_gradient', self._use_gradient)
		self.add_tool_action_boolean('pencil-outline', self._use_outline)
		self.add_tool_action_boolean('line-ortholock', self._ortholock)
		self._set_options_attributes() # Not optimal but more readable

	def get_tooltip(self, event_x, event_y, motion_behavior):
		if motion_behavior != 1:
			return None # no line is being drawn

		delta_x = abs(self.x_press - event_x)
		delta_y = abs(self.y_press - event_y)
		line1 = _("Width: %spx") % str(delta_x)
		line2 = _("Height: %spx") % str(delta_y)
		length = round(math.sqrt(delta_x * delta_x + delta_y * delta_y), 2)
		return line1 + "\n" + line2 + "\n" + _("Length: %spx") % str(length)

	############################################################################

	def _set_active_shape(self):
		state_as_string = self.get_option_value('line_shape')
		if state_as_string == 'thin':
			self._cap_id = cairo.LineCap.BUTT
		else:
			self._cap_id = cairo.LineCap.ROUND

	def get_options_label(self):
		return _("Line options")

	def _set_options_attributes(self):
		self._use_outline = self.get_option_value('pencil-outline')
		self._dashes_type = self.get_option_value('dashes-type')
		self._arrow_type = self.get_option_value('arrow-type')
		self._use_gradient = self.get_option_value('use_gradient')
		self._ortholock = self.get_option_value('line-ortholock')
		self._set_active_shape()

	def get_editing_tips(self):
		self._set_options_attributes()
		is_arrow = self._arrow_type != 'none'
		use_dashes = self._dashes_type != 'none'

		label_options = self.label
		if is_arrow and use_dashes:
			label_options += " - " + _("Dashed arrow")
		elif is_arrow:
			label_options += " - " + _("Arrow")
		elif use_dashes:
			label_options += " - " + _("Dashed")
		else:
			label_options = None

		label_modifier_shift = self.label + " - "
		if self._ortholock:
			label_modifier_shift += _("Press <Shift> to unlock the line direction")
		else:
			label_modifier_shift += _("Press <Shift> to lock the line direction")

		label_modifier_alt = self.label + " - " + \
		                         _("Press <Alt> to toggle the 'outline' option")

		if self.get_image().get_mouse_is_pressed():
			label_modifier_shift = None
			label_modifier_alt = None

		full_list = [label_options, label_modifier_shift, label_modifier_alt]
		return list(filter(None, full_list))

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.set_common_values(event.button, event_x, event_y)

		self.update_modifier_state(event.state)
		if 'SHIFT' in self._modifier_keys:
			self._ortholock = not self._ortholock
		if 'ALT' in self._modifier_keys:
			self._use_outline = not self._use_outline

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		if not render:
			return
		operation = self.build_operation(event_x, event_y)
		self.do_tool_operation(operation)

	def on_release_on_area(self, event, surface, event_x, event_y):
		operation = self.build_operation(event_x, event_y)
		self.apply_operation(operation)

	############################################################################

	def build_operation(self, event_x, event_y):
		operation = {
			'tool_id': self.id,
			'rgba': self.main_color,
			'rgba2': self.secondary_color,
			'antialias': self._use_antialias,
			'operator': self._operator,
			'line_width': self.tool_width,
			'line_cap': self._cap_id,
			'dashes': self._dashes_type,
			'arrow': self._arrow_type,
			'gradient': self._use_gradient,
			'outline': self._use_outline,
			'ortholock': self._ortholock,
			'x_release': event_x,
			'y_release': event_y,
			'x_press': self.x_press,
			'y_press': self.y_press
		}
		return operation

	def do_tool_operation(self, operation):
		cairo_context = self.start_tool_operation(operation)
		cairo_context.set_operator(operation['operator'])
		line_width = operation['line_width']
		cairo_context.set_line_width(line_width)
		x1 = operation['x_press']
		y1 = operation['y_press']
		x2 = operation['x_release']
		y2 = operation['y_release']

		if operation['ortholock']:
			delta_x = abs(x1 - x2)
			delta_y = abs(y1 - y2)
			if delta_x > 2 * delta_y:
				# Strictly horizontal line
				y2 = y1
			elif delta_x * 2 < delta_y:
				# Strictly vertical line
				x2 = x1
			else:
				# 45° line
				delta45 = min(delta_x, delta_y)
				x2 = x1 + delta45 if (x1 - x2 < 0) else x1 - delta45
				y2 = y1 + delta45 if (y1 - y2 < 0) else y1 - delta45

		self.set_dashes_and_cap(cairo_context, line_width, \
		                             operation['dashes'], operation['line_cap'])

		if operation['arrow'] == 'double':
			utilities_add_arrow_triangle(cairo_context, x1, y1, x2, y2, line_width)

		# We don't memorize the path because all coords are here anyway for the
		# linear gradient and/or the arrow.
		cairo_context.move_to(x1, y1)
		cairo_context.line_to(x2, y2)

		if operation['arrow'] != 'none':
			utilities_add_arrow_triangle(cairo_context, x2, y2, x1, y1, line_width)

		if operation['outline']:
			cairo_context.set_source_rgba(*operation['rgba2'])
			cairo_context.set_line_width(line_width * 1.2 + 2)
			cairo_context.stroke_preserve()

		if operation['gradient']:
			pattern = cairo.LinearGradient(x1, y1, x2, y2)
			pattern.add_color_stop_rgba(0.1, *operation['rgba'])
			pattern.add_color_stop_rgba(0.9, *operation['rgba2'])
			cairo_context.set_source(pattern)
		else:
			cairo_context.set_source_rgba(*operation['rgba'])
		cairo_context.set_line_width(line_width)
		cairo_context.stroke()

	############################################################################
################################################################################




# tool_brush.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk
from .abstract_classic_tool import AbstractClassicTool

from .brush_simple import BrushSimple
from .brush_airbrush import BrushAirbrush
from .brush_nib import BrushNib
from .brush_hairy import BrushHairy

class ToolBrush(AbstractClassicTool):
	__gtype_name__ = 'ToolBrush'

	def __init__(self, window, **kwargs):
		super().__init__('brush', _("Brush"), 'tool-brush-symbolic', window)
		self.use_operator = True
		self._used_pressure = False

		self._brushes_dict = {
			'simple': BrushSimple('simple', self),
			'airbrush': BrushAirbrush('airbrush', self),
			'calligraphic': BrushNib('calligraphic', self),
			'hairy': BrushHairy('hairy', self),
		}

		self._brush_type = 'simple'
		self._brush_dir = 'right'
		self.add_tool_action_enum('brush-type', self._brush_type)
		self.add_tool_action_enum('brush-dir', self._brush_dir)

	def get_options_label(self):
		return _("Brush options")

	def get_editing_tips(self):
		self._brush_type = self.get_option_value('brush-type')
		self._brush_dir = self.get_option_value('brush-dir')

		enable_direction = self._brush_type == 'calligraphic'
		self.set_action_sensitivity('brush-dir', enable_direction)

		active_brush = self._brushes_dict[self._brush_type]
		return active_brush._get_tips(self._used_pressure, self._brush_dir)

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.set_common_values(event.button, event_x, event_y)
		self._manual_path = []
		self._add_pressured_point(event_x, event_y, event)
		self._used_pressure = self._manual_path[0]['p'] is not None

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		self._add_pressured_point(event_x, event_y, event)
		if render:
			operation = self.build_operation()
			self.do_tool_operation(operation)

	def on_release_on_area(self, event, surface, event_x, event_y):
		self._add_pressured_point(event_x, event_y, event)
		operation = self.build_operation()
		operation['is_preview'] = False
		self.apply_operation(operation)

	############################################################################

	def _add_pressured_point(self, event_x, event_y, event):
		new_point = {
			'x': event_x,
			'y': event_y,
			'p': self._get_pressure(event)
		}
		self._manual_path.append(new_point)

	def _get_pressure(self, event):
		device = event.get_source_device()
		# print(device)
		if device is None:
			return None
		# source = device.get_source()
		# print(source) # J'ignore s'il faut faire quelque chose de cette info

		tool = event.get_device_tool()
		# print(tool) # ça indique qu'on a ici un appareil dédié au dessin (vaut
		# `None` si c'est pas le cas). Autrement on peut avoir des valeurs comme
		# Gdk.DeviceToolType.PEN, .ERASER, .BRUSH, .PENCIL, ou .AIRBRUSH, et
		# aussi (même si jsuis pas sûr ce soit pertinent) .UNKNOWN, .MOUSE et
		# .LENS, on pourrait adapter le comportement (couleur/opérateur/etc.)
		# à cette information à l'avenir.

		pressure = event.get_axis(Gdk.AxisUse.PRESSURE)
		# print(pressure)
		if pressure is None:
			return None
		return pressure

	############################################################################

	def build_operation(self):
		operation = {
			'tool_id': self.id,
			'brush_id': self._brush_type,
			'nib_dir': self._brush_dir,
			'rgba': self.main_color,
			'operator': self._operator,
			'line_width': self.tool_width,
			'antialias': self._use_antialias,
			'is_preview': True,
			'smooth': not self.get_image().is_zoomed_surface_sharp(),
			'path': self._manual_path
		}
		return operation

	def do_tool_operation(self, operation):
		if operation['path'] is None or len(operation['path']) < 1:
			return
		cairo_context = self.start_tool_operation(operation)

		active_brush = self._brushes_dict[operation['brush_id']]
		active_brush.do_brush_operation(cairo_context, operation)

	############################################################################
################################################################################




# optionsbar_scale.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
from .abstract_optionsbar import AbstractOptionsBar
from .utilities_units import utilities_add_unit_to_spinbtn

class OptionsBarScale(AbstractOptionsBar):
	__gtype_name__ = 'OptionsBarScale'

	def __init__(self, scale_tool):
		super().__init__()

		self._build_ui('transform/abstract-optionsbar-transform.ui')
		builder = self._hydrate_transform_tool('transform/optionsbar-scale.ui')

		self.width_btn = builder.get_object('width_btn')
		self.height_btn = builder.get_object('height_btn')
		utilities_add_unit_to_spinbtn(self.height_btn, 4, 'px')
		utilities_add_unit_to_spinbtn(self.width_btn, 4, 'px')

		self.width_100_btn = builder.get_object('width_100_btn')
		self.height_100_btn = builder.get_object('height_100_btn')
		utilities_add_unit_to_spinbtn(self.height_100_btn, 3, '%')
		utilities_add_unit_to_spinbtn(self.width_100_btn, 3, '%')

		self.width_label = builder.get_object('width_label')
		self.height_label = builder.get_object('height_label')
		self.separator = builder.get_object('separator')

	def init_adaptability(self):
		super().init_adaptability()
		# this ^ just did a show_all, but both sets of spinbuttons can't be
		# visible at the same time: the limit size only depends on one of them
		self.width_100_btn.set_visible(False)
		self.height_100_btn.set_visible(False)
		temp_limit_size = self.centered_box.get_preferred_width()[0] + \
		                    self.cancel_btn.get_preferred_width()[0] + \
		                   self.options_btn.get_preferred_width()[0] + \
		                      self.help_btn.get_preferred_width()[0] + \
		                     self.apply_btn.get_preferred_width()[0]
		self._set_limit_size(temp_limit_size)

	def set_compact(self, state):
		super().set_compact(state)
		if state:
			self.centered_box.set_orientation(Gtk.Orientation.VERTICAL)
		else:
			self.centered_box.set_orientation(Gtk.Orientation.HORIZONTAL)
		# future possible improvement: the pane may not compact the same way if
		# the manipulation concerns the selection.
		self.width_label.set_visible(not state)
		self.height_label.set_visible(not state)
		self.separator.set_visible(not state)

	############################################################################
################################################################################




# properties.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GdkPixbuf, Pango

class DrPropertiesDialog(Gtk.Dialog):
	__gtype_name__ = 'DrPropertiesDialog'

	units_dict = {
		'px': _("%s px"),
		'cm': _("%s cm"),
		'in': _("%s in")
	}

	def __init__(self, parent_window, image):
		wants_csd = 'h' in parent_window.deco_layout
		super().__init__(use_header_bar=wants_csd, \
		                 destroy_with_parent=True, \
		                 transient_for=parent_window, \
		                 title=_("Image properties"))
		self._image = image
		self._build_ui()
		if wants_csd:
			subtitle = self._image.get_filename_for_display()
			self.get_titlebar().set_subtitle(subtitle)
			# put stackswitcher [Image|Metadata] as title_widget instead
		self.set_default_size(400, 150)
		self.show()

	def _build_ui(self):
		"""Fill the dialog with labels displaying correct information."""
		ui_file_path = '/com/github/maoschanz/drawing/ui/properties.ui'
		builder = Gtk.Builder.new_from_resource(ui_file_path)
		self.get_content_area().add(builder.get_object('main_widget'))
		self._grid = builder.get_object('image_props_grid')

		# Size (and size unit) #################################################

		self.label_width = builder.get_object('label_width')
		self.label_height = builder.get_object('label_height')

		self.unit = 'px'
		btn_px = builder.get_object('units_pixels')
		btn_in = builder.get_object('units_inches')
		btn_cm = builder.get_object('units_cm')
		btn_in.join_group(btn_px)
		btn_cm.join_group(btn_px)
		btn_px.connect('toggled', self._set_unit, 'px')
		btn_cm.connect('toggled', self._set_unit, 'cm')
		btn_in.connect('toggled', self._set_unit, 'in')

		self._set_size_labels()

		# Path, format and colorspace ##########################################

		if self._image.gfile is None:
			label_path = _("Unsaved file")
			label_file_format = _("Unsaved file")
		else:
			label_path = self._image.get_file_path()
			(pb_format, w, h) = GdkPixbuf.Pixbuf.get_file_info(label_path)
			label_file_format = pb_format.get_name()

		# Context: the path of the edited file
		self._add_grid_row(0, _("Path"), label_path)
		# Context: the file format of the edited file
		self._add_grid_row(1, _("Format"), label_file_format)
		self._add_grid_row(2, _("Colorspace"), self._set_colorspace_label())
		# TODO display both the colorspace of the file and of the surface, with
		# a warning if there might be a loss of data

	def _add_grid_row(self, index, key, value):
		"""Adds a row 2 labels (a key and a value) to the dialog's main grid."""
		key_label = Gtk.Label(label=key, halign=Gtk.Align.END, visible=True)
		self._grid.attach(key_label, 0, index, 1, 1)
		value_label = Gtk.Label(label=value, halign=Gtk.Align.START, \
		                 wrap=True, wrap_mode=Pango.WrapMode.CHAR, visible=True)
		value_label.get_style_context().add_class('dim-label')
		self._grid.attach(value_label, 1, index, 2, 1)

	def _set_colorspace_label(self):
		# Useless, it will always return "ARGB32"
		enum = {
			0: "ARGB32",
			1: "RGB24",
			2: "A8",
			3: "A1",
			4: "RGB16_565",
			5: "RGB30",
		}
		cairo_format = self._image.get_surface().get_format()
		# Context: an invalid colorspace format
		colorspace_text = enum.get(cairo_format, _("Invalid format"))
		return colorspace_text

	############################################################################

	def _set_size_labels(self):
		"""Set the labels for picture width and height according to the selected
		unit (px, cm or in)."""
		# TODO contrast this size with the size of the saved file
		px_width = self._image.get_pixbuf_width()
		px_height = self._image.get_pixbuf_height()
		if self.unit == 'in':
			width = round(px_width/96.0, 2)
			height = round(px_height/96.0, 2)
		elif self.unit == 'cm':
			width = round(px_width/37.795275591, 2)
			height = round(px_height/37.795275591, 2)
		else:
			width = px_width
			height = px_height
		self.label_width.set_label(self.units_dict[self.unit] % str(width))
		self.label_height.set_label(self.units_dict[self.unit] % str(height))

	def _set_unit(self, *args):
		self.unit = args[1]
		self._set_size_labels()

	############################################################################
################################################################################

# TODO pixbuf.get_options() ⇒
# * The PNG loader provides the tEXt ancillary chunk (http://www.libpng.org/pub/png/spec/1.2/PNG-Chunks.html 4.2.3) key/value pairs as options :

# Title            Short (one line) title or caption for image
# Author           Name of image's creator
# Description      Description of image (possibly long)
# Copyright        Copyright notice
# Creation Time    Time of original image creation
# Software         Software used to create the image
# Disclaimer       Legal disclaimer
# Warning          Warning of nature of content
# Source           Device used to create the image
# Comment          Miscellaneous comment; conversion from GIF comment

# * the TIFF and JPEG loaders return an “orientation” option string that corresponds to the embedded TIFF/Exif orientation tag (if present).
# * the TIFF loader sets the “multipage” option string to “yes” when a multi-page TIFF is loaded.
# * the JPEG and PNG loaders set “x-dpi” and “y-dpi” if the file contains image density information in dots per inch.
# * the JPEG loader sets the “comment” option with the comment EXIF tag.




# tool_skew.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk
from .abstract_transform_tool import AbstractCanvasTool
from .optionsbar_skew import OptionsBarSkew
from .utilities_overlay import utilities_show_handles_on_context

class ToolSkew(AbstractCanvasTool):
	__gtype_name__ = 'ToolSkew'

	def __init__(self, window):
		# This is the name of the tool changing rectangles into parallelograms.
		# It's synonymous with tilt, slant, bend. If you need a reference to
		# translate it, this is named after MS Paint's "Stretch/Skew" dialog.
		super().__init__('skew', _("Skew"), 'tool-skew-symbolic', window)
		self._x = 0
		self._y = 0
		self.add_tool_action_simple('skew-exists', self.select_flowbox_child)
		self.add_tool_action_enum('crop-expand', 'initial')

	def try_build_pane(self):
		self.pane_id = 'skew'
		self.window.options_manager.try_add_bottom_pane(self.pane_id, self)

	def build_bottom_pane(self):
		bar = OptionsBarSkew()
		self.yx_spinbtn = bar.yx_spinbtn
		self.xy_spinbtn = bar.xy_spinbtn
		self.yx_spinbtn.connect('value-changed', self.on_coord_changed)
		self.xy_spinbtn.connect('value-changed', self.on_coord_changed)
		return bar

	def get_options_model(self):
		if self.apply_to_selection:
			return None
		else:
			return super().get_options_model()

	def get_options_label(self):
		if self.apply_to_selection:
			return super().get_options_label()
		else:
			return _("Skewing options")

	def get_editing_tips(self):
		label_direction = _("The directions of the deformation are hinted by the mouse pointer")

		label_modifiers = None
		if self.apply_to_selection:
			label_action = _("Skewing the selection")
			label_confirm = None
		else:
			label_action = _("Skewing the canvas")
			label_confirm = self.label + " - " + \
			                         _("Don't forget to confirm the operation!")
			if not self.get_image().get_mouse_is_pressed():
				label_modifiers = _("Press <Alt>, <Shift>, or both, to " + \
				                      "quickly change the 'expand with' option")

		full_list = [label_action, label_direction, label_confirm, label_modifiers]
		return list(filter(None, full_list))

	def on_tool_selected(self, *args):
		super().on_tool_selected()
		self.set_action_sensitivity('crop-expand', not self.apply_to_selection)
		self._reset_values()

	############################################################################

	def get_yx(self):
		return self.yx_spinbtn.get_value_as_int()

	def get_xy(self):
		return self.xy_spinbtn.get_value_as_int()

	def _set_cursor_name(self):
		if self._directions == '':
			self.cursor_name = 'not-allowed'
			return
		if self._directions == 'nw' or self._directions == 'se':
			self.cursor_name = 'nwse-resize'
		elif self._directions == 'ne' or self._directions == 'sw':
			self.cursor_name = 'nesw-resize'
		elif self._directions == 'n' or self._directions == 's':
			self.cursor_name = 'ew-resize'
		elif self._directions == 'e' or self._directions == 'w':
			self.cursor_name = 'ns-resize'

	def _get_width(self):
		if self.apply_to_selection:
			source_pixbuf = self.get_selection_pixbuf()
		else:
			source_pixbuf = self.get_main_pixbuf()
		return source_pixbuf.get_width()

	def _get_height(self):
		if self.apply_to_selection:
			source_pixbuf = self.get_selection_pixbuf()
		else:
			source_pixbuf = self.get_main_pixbuf()
		return source_pixbuf.get_height()

	def on_coord_changed(self, *args):
		self.build_and_do_op()

	def _reset_values(self, *args):
		self._update_expansion_rgba()
		self._yx = 0 # vertical deformation
		self._xy = 0 # horizontal deformation
		self.yx_spinbtn.set_value(0)
		self.xy_spinbtn.set_value(0)
		self.build_and_do_op()

	############################################################################

	def on_unclicked_motion_on_area(self, event, surface):
		self.set_directional_cursor(event.x, event.y)

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.x_press = event_x
		self.y_press = event_y
		self._yx = self.get_yx() # vertical deformation
		self._xy = self.get_xy() # horizontal deformation
		# TODO répliquer ce que fait le scale avec son x2/y2 qui évite un effet
		# flamby dégueulasse lié aux arrondis ?
		self._update_expansion_rgba(event.button)

		self.update_modifier_state(event.state)
		if 'SHIFT' in self._modifier_keys and 'ALT' in self._modifier_keys:
			self._force_expansion_rgba('secondary', event.button)
		elif 'SHIFT' in self._modifier_keys:
			self._force_expansion_rgba('alpha')
		elif 'ALT' in self._modifier_keys:
			self._force_expansion_rgba('initial')

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		if self._directions == '' or not render:
			return

		delta_x = event_x - self.x_press
		delta_y = event_y - self.y_press

		yx = self._yx
		xy = self._xy
		if 'n' in self._directions:
			xy -= 100 * delta_x/self._get_width()
		if 's' in self._directions:
			xy += 100 * delta_x/self._get_width()
		if 'w' in self._directions:
			yx -= 100 * delta_y/self._get_height()
		if 'e' in self._directions:
			yx += 100 * delta_y/self._get_height()

		self.yx_spinbtn.set_value(yx)
		self.xy_spinbtn.set_value(xy)

	def on_release_on_area(self, event, surface, event_x, event_y):
		self.on_motion_on_area(event, surface, event_x, event_y)
		self.build_and_do_op() # technically already done

	############################################################################

	def on_draw_above(self, area, cairo_context):
		x1 = 0
		y1 = 0
		scaled_xy = abs(self.get_xy()) * (self._get_height() /  self._get_width())
		scaled_yx = abs(self.get_yx()) * (self._get_width() /  self._get_height())
		p_xy = (scaled_xy + 100) / 100
		p_yx = (scaled_yx + 100) / 100
		x2 = x1 + self._get_width() * p_xy
		y2 = y1 + self._get_height() * p_yx

		x1, x2, y1, y2 = self.get_image().get_corrected_coords(x1, x2, y1, y2, \
		                                         self.apply_to_selection, False)
		self._draw_temp_pixbuf(cairo_context, x1, y1)
		thickness = self.get_overlay_thickness()
		utilities_show_handles_on_context(cairo_context, x1, x2, y1, y2, thickness)

	############################################################################

	def build_operation(self):
		operation = {
			'tool_id': self.id,
			'is_selection': self.apply_to_selection,
			'is_preview': True,
			'local_dx': 0,
			'local_dy': 0,
			'yx': self.yx_spinbtn.get_value_as_int()/100,
			'xy': self.xy_spinbtn.get_value_as_int()/100,
			'rgba': self._expansion_rgba,
		}
		return operation

	def do_tool_operation(self, operation):
		self.start_tool_operation(operation)
		if operation['is_selection']:
			source_pixbuf = self.get_selection_pixbuf()
			prefill = False
		else:
			source_pixbuf = self.get_main_pixbuf()
			prefill = True
		source_surface = Gdk.cairo_surface_create_from_pixbuf(source_pixbuf, 0, None)
		source_surface.set_device_scale(self.scale_factor(), self.scale_factor())

		xy = operation['xy']
		x0 = 0.0
		if xy < 0:
			x0 = int(-1 * xy * source_surface.get_height())
		yx = operation['yx']
		y0 = 0.0
		if yx < 0:
			y0 = int(-1 * yx * source_surface.get_width())
		coefs = [1.0, yx, xy, 1.0, x0, y0]

		new_surface = self.get_deformed_surface(source_surface, coefs, prefill)
		new_pixbuf = Gdk.pixbuf_get_from_surface(new_surface, 0, 0, \
		                      new_surface.get_width(), new_surface.get_height())
		self.get_image().set_temp_pixbuf(new_pixbuf)
		self.common_end_operation(operation)

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo, math
from gi.repository import Gdk, GdkPixbuf
from .message_dialog import DrMessageDialog

################################################################################

def utilities_get_rgba_for_xy(surface, x, y):
	# Guard clause: we can't perform color picking outside of the surface
	if x < 0 or x > surface.get_width() or y < 0 or y > surface.get_height():
		return None
	screenshot = Gdk.pixbuf_get_from_surface(surface, float(x), float(y), 1, 1)
	rgba_vals = screenshot.get_pixels()
	return rgba_vals

def utilities_get_magic_path(surface, x, y, window, coef):
	"""This method tries to build a path defining an area of the same color. It
	will mainly be used to paint this area, or to select it."""
	cairo_context = cairo.Context(surface)
	old_color = utilities_get_rgba_for_xy(surface, x, y)

	# Cairo doesn't provide methods for what we want to do. I will have to
	# define myself how to decide what should be filled.
	# The heuristic here is that we create a hull containing the area of
	# color we want to paint. We don't care about "enclaves" of other colors.
	while (utilities_get_rgba_for_xy(surface, x, y) == old_color) and y > 0:
		y = y - 1
	y = y + 1 # sinon ça crashe ?
	cairo_context.move_to(x, y)
	(first_x, first_y) = (x, y)

	# 0 1 2
	# 7   3
	# 6 5 4

	direction = 5
	should_stop = False
	i = 0

	x_shift = [-1 * coef, 0, coef, coef, coef, 0, -1 * coef, -1 * coef]
	y_shift = [-1 * coef, -1 * coef, -1 * coef, 0, coef, coef, coef, 0]

	while (not should_stop and i < 50000):
		new_x = -10
		new_y = -10
		end_circle = False
		j = 0
		while (not end_circle) or (j < 8):
			future_x = x+x_shift[direction]
			future_y = y+y_shift[direction]
			if (utilities_get_rgba_for_xy(surface, future_x, future_y) == old_color) \
			and (future_x > 0) and (future_y > 0) \
			and (future_x < surface.get_width()) \
			and (future_y < surface.get_height()-2): # ???
				new_x = future_x
				new_y = future_y
				direction = (direction+1) % 8
			elif (x != new_x or y != new_y):
				x = new_x+x_shift[direction]
				y = new_y+y_shift[direction]
				end_circle = True
			# else:
			# 	print('cas emmerdant')
			j = j+1

		direction = (direction+4) % 8
		# print('direction :', direction)
		if (new_x != -10):
			cairo_context.line_to(x, y)
		# else:
		#	 print("TENTATIVE ABUSIVE D'AJOUT")
		#	 should_stop = True

		if (i > 10) and (first_x-5 < x < first_x+5) and (first_y-5 < y < first_y+5):
			should_stop = True

		i = i + 1
		if i == 2000:
			dialog, continue_id = launch_infinite_loop_dialog(window)
			result = dialog.run()
			if result == continue_id: # Continue
				dialog.destroy()
			else: # Cancel
				dialog.destroy()
				return

	cairo_context.close_path()
	# print('i: ' + str(i))
	return cairo_context.copy_path()

def launch_infinite_loop_dialog(window):
	dialog = DrMessageDialog(window)
	cancel_id = dialog.set_action(_("Cancel"), None)
	continue_id = dialog.set_action(_("Continue"), None, True)
	dialog.add_string(_("The area seems poorly delimited, or is very complex."))
	dialog.add_string(_("This algorithm may not be able to manage the wanted area."))
	dialog.add_string(_("Do you want to abort the operation, or to let the tool struggle ?"))
	return dialog, continue_id

################################################################################

# The coordinates of the corners of the triangle. These points are defined as if
# the end of the arrow line is at "0, 0" and rotated by 0 degrees.
ARROW_TRIANGLE = [
	(0.0, 0.0),
	(-2.747, 1.0),
	(-2.747, -1.0),
]

MIN_ARROW_SCALE = 3

def utilities_add_arrow_triangle(cairo_context, x2, y2, x1, y1, line_width):
	"""Adds a triangular head to the current path."""
	if x1 == x2 and y1 == y2:
		line_angle = 0
	else:
		line_angle = math.atan2(y2 - y1, x2 - x1)
	sin, cos = math.sin(line_angle), math.cos(line_angle)

	# FIXME cases with very short last segment
	# if scale == line_width, dashed arrow will look like shit
	scale = max(line_width * 1.1, MIN_ARROW_SCALE)

	# when the last segment of the line is very short, the head is scaled down
	# last_segment_length = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
	# scale = min(scale, last_segment_length/line_width)
	# ^XXX broken with dashes
	# TODO ça peut ne plus être nécessaire si je décale la tête

	head = []
	# scale, rotate and translate the arrow triangle
	for x, y in ARROW_TRIANGLE:
		p = (x * scale, y * scale)
		p = (p[0] * cos - p[1] * sin, p[0] * sin + p[1] * cos)
		p = (p[0] + x2, p[1] + y2)
		head.append(p)

	# draw the arrow triangle
	_draw_head(cairo_context, head, True)

	# shameful hack to get a full head in cases where the line is dashed
	if cairo_context.get_dash_count() > 0:
		_draw_head(cairo_context, head, False)
		_draw_head(cairo_context, head, False)
		_draw_head(cairo_context, head, False)
		_draw_head(cairo_context, head, False)

	# XXX the path isn't filled because the path is opened by the first point of
	# the line/curve and it can't be closed so easily
	cairo_context.close_path()
	# The stroke must be done afterwards, by the calling method

def _draw_head(cairo_context, head, first_try):
	if first_try:
		cairo_context.move_to(*head[0])
	else:
		cairo_context.line_to(*head[0])
	cairo_context.line_to(*head[1])
	cairo_context.line_to(*head[2])

################################################################################
# Path smoothing ###############################################################

def utilities_smooth_path(cairo_context, cairo_path):
	"""Extrapolate a path made of straight lines into a path made of curves. New
	points are added according to the length of the line it replaces, the length
	of the previous one, and the length of the next one."""
	x1 = y1 = None
	x2 = y2 = None
	x3 = y3 = None
	x4 = y4 = None
	for pts in cairo_path:
		if pts[1] == ():
			continue
		x1, y1, x2, y2, x3, y3, x4, y4 = _next_arc(cairo_context, \
		                           x2, y2, x3, y3, x4, y4, pts[1][0], pts[1][1])
	_next_arc(cairo_context, x2, y2, x3, y3, x4, y4, None, None)
	# cairo_context.stroke()

def _next_point(x1, y1, x2, y2, dist):
	coef = 0.1
	dx = x2 - x1
	dy = y2 - y1
	angle = math.atan2(dy, dx)
	nx = x2 + math.cos(angle) * dist * coef
	ny = y2 + math.sin(angle) * dist * coef
	return nx, ny

def _next_arc(cairo_context, x1, y1, x2, y2, x3, y3, x4, y4):
	if x2 is None or x3 is None:
		# No drawing possible yet, just continue to the next point
		return x1, y1, x2, y2, x3, y3, x4, y4
	dist = math.sqrt( (x2 - x3) * (x2 - x3) + (y2 - y3) * (y2 - y3) )
	if x1 is None and x4 is None:
		cairo_context.move_to(x2, y2)
		cairo_context.line_to(x3, y3)
		return x1, y1, x2, y2, x3, y3, x4, y4
	elif x1 is None:
		nx1, ny1 = x2, y2
		nx2, ny2 = _next_point(x4, y4, x3, y3, dist)
	elif x4 is None:
		nx1, ny1 = _next_point(x1, y1, x2, y2, dist)
		nx2, ny2 = x3, y3
	else:
		nx1, ny1 = _next_point(x1, y1, x2, y2, dist)
		nx2, ny2 = _next_point(x4, y4, x3, y3, dist)
	cairo_context.curve_to(nx1, ny1, nx2, ny2, x3, y3)
	return x1, y1, x2, y2, x3, y3, x4, y4

################################################################################




# new_image_dialog.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gdk
from .utilities_units import utilities_add_unit_to_spinbtn

class DrCustomImageDialog(Gtk.Dialog):
	__gtype_name__ = 'DrCustomImageDialog'

	def __init__(self, appwindow):
		wants_csd = 'h' in appwindow.deco_layout
		super().__init__(use_header_bar=wants_csd, destroy_with_parent=True, \
		         transient_for=appwindow, title=_("New Image With Custom Size"))
		self._app_settings = appwindow.gsettings
		self._build_ui()
		self.set_default_size(450, 200)

	def _build_ui(self):
		self.add_button(_("Cancel"), Gtk.ResponseType.CANCEL)
		# Context: Create a new image
		btn = self.add_button(_("Create"), Gtk.ResponseType.OK)
		btn.get_style_context().add_class('suggested-action')

		resource_path = '/com/github/maoschanz/drawing/ui/new-image-dialog.ui'
		builder = Gtk.Builder.new_from_resource(resource_path)
		props_content_area = builder.get_object('props_content_area')
		self.get_content_area().add(props_content_area)

		self._width_btn = builder.get_object('spin_width')
		utilities_add_unit_to_spinbtn(self._width_btn, 4, 'px')
		default_w = self._app_settings.get_int('default-width')
		self._width_btn.set_value(default_w)

		self._height_btn = builder.get_object('spin_height')
		utilities_add_unit_to_spinbtn(self._height_btn, 4, 'px')
		default_h = self._app_settings.get_int('default-height')
		self._height_btn.set_value(default_h)

		self._color_btn = builder.get_object('color_btn')
		background_rgba = self._app_settings.get_strv('default-rgba')
		r = float(background_rgba[0])
		g = float(background_rgba[1])
		b = float(background_rgba[2])
		a = float(background_rgba[3])
		color = Gdk.RGBA(red=r, green=g, blue=b, alpha=a)
		self._color_btn.set_rgba(color)

		self._default_checkbtn = builder.get_object('default_checkbtn')

	############################################################################

	def get_values(self):
		"""Will be called from the main window if the user clicks on "Create",
		and returns the values set by the user."""
		width = self._width_btn.get_value_as_int()
		height = self._height_btn.get_value_as_int()
		rgba = self._color_btn.get_rgba()
		rgba = [str(rgba.red), str(rgba.green), str(rgba.blue), str(rgba.alpha)]
		if self._default_checkbtn.get_active():
			self._app_settings.set_int('default-width', width)
			self._app_settings.set_int('default-height', height)
			self._app_settings.set_strv('default-rgba', rgba)
		return width, height, rgba

	############################################################################
################################################################################



# tool_experiment.py

import cairo, math, random
from gi.repository import Gdk
from .abstract_classic_tool import AbstractClassicTool
from .utilities_paths import utilities_smooth_path

class ToolExperiment(AbstractClassicTool):
	__gtype_name__ = 'ToolExperiment'

	def __init__(self, window, **kwargs):
		super().__init__('experiment', _("Experiment"), 'applications-utilities-symbolic', window)

		# In order to draw pressure-sensitive lines, the path is collected as
		# an array whose elements are dicts (keys are 'x', 'y', 'p'). An actual
		# cairo path will be built from this data when necessary.
		self._manual_path = []

		self._use_antialias = True
		self._operators_dict = {
			'CLEAR': cairo.Operator.CLEAR,
			'SOURCE': cairo.Operator.SOURCE,
			'OVER': cairo.Operator.OVER,
			'IN': cairo.Operator.IN,
			'OUT': cairo.Operator.OUT,
			'ATOP': cairo.Operator.ATOP,
			'DEST': cairo.Operator.DEST,
			'DEST_OVER': cairo.Operator.DEST_OVER,
			'DEST_IN': cairo.Operator.DEST_IN,
			'DEST_OUT': cairo.Operator.DEST_OUT,
			'DEST_ATOP': cairo.Operator.DEST_ATOP,
			'XOR': cairo.Operator.XOR,
			'ADD': cairo.Operator.ADD,
			'SATURATE': cairo.Operator.SATURATE,
			'MULTIPLY': cairo.Operator.MULTIPLY,
			'SCREEN': cairo.Operator.SCREEN,
			'OVERLAY': cairo.Operator.OVERLAY,
			'DARKEN': cairo.Operator.DARKEN,
			'LIGHTEN': cairo.Operator.LIGHTEN,
			'COLOR_DODGE': cairo.Operator.COLOR_DODGE,
			'COLOR_BURN': cairo.Operator.COLOR_BURN,
			'HARD_LIGHT': cairo.Operator.HARD_LIGHT,
			'SOFT_LIGHT': cairo.Operator.SOFT_LIGHT,
			'DIFFERENCE': cairo.Operator.DIFFERENCE,
			'EXCLUSION': cairo.Operator.EXCLUSION,
			'HSL_HUE': cairo.Operator.HSL_HUE,
			'HSL_SATURATION': cairo.Operator.HSL_SATURATION,
			'HSL_COLOR': cairo.Operator.HSL_COLOR,
			'HSL_LUMINOSITY': cairo.Operator.HSL_LUMINOSITY
		}
		self._operator_label = "OVER"
		self.operator2 = cairo.Operator.OVER
		self._selected_mode = 'feather'
		self._feather_dir = 'up'

		self.add_tool_action_enum('experiment_operator', self._operator_label)
		self.add_tool_action_enum('experiment_mode', self._selected_mode)

	def build_flowbox_child(self, flowbox):
		super().build_flowbox_child(flowbox)
		self._label_box.get_style_context().add_class('dim-label')

	def get_editing_tips(self):
		return [
			self.label + " - " + self.get_options_label(),
			"You're not supposed to use this tool (development only)."
		]

	def get_options_label(self):
		self._set_active_operator()
		self._set_active_mode()
		if self._selected_mode == 'simple':
			return self._operator_label
		else:
			return self._selected_mode

	############################################################################

	def _set_active_mode(self, *args):
		state_as_string = self.get_option_value('experiment_mode')
		self._selected_mode = state_as_string

	def _set_active_operator(self, *args):
		state_as_string = self.get_option_value('experiment_operator')
		self.operator2 = self._operators_dict[state_as_string]
		self._operator_label = state_as_string

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.get_options_label()
		self.x_press = event_x
		self.y_press = event_y
		self.set_common_values(event.button, event_x, event_y)
		self._manual_path = []
		self._add_pressured_point(event_x, event_y, event)

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		self._add_pressured_point(event_x, event_y, event)
		if render:
			operation = self.build_operation()
			self.do_tool_operation(operation)

	def on_release_on_area(self, event, surface, event_x, event_y):
		self._add_pressured_point(event_x, event_y, event)
		operation = self.build_operation()
		operation['is_preview'] = False
		self.apply_operation(operation)

	############################################################################

	def _add_pressured_point(self, event_x, event_y, event):
		new_point = {
			'x': event_x,
			'y': event_y,
			'p': self._get_pressure(event)
		}
		self._manual_path.append(new_point)

	def _get_pressure(self, event):
		device = event.get_source_device()
		# print(device)
		if device is None:
			return None
		source = device.get_source()
		# print(source) # TODO ça peut être Gdk.InputSource.MOUSE, ou .TOUCHPAD,
		# ou bien des trucs pertinents comme Gdk.InputSource.ERASER ou .PEN
		# J'ignore s'il faut faire quelque chose de cette information ? Il y a
		# ptêt des touchpads ou des touchscreens sensibles à la pression non ?

		tool = event.get_device_tool()
		# print(tool) # TODO ça indique qu'on a ici un appareil dédié au dessin
		# (vaut `None` si c'est pas le cas). Autrement on peut avoir des valeurs
		# comme Gdk.DeviceToolType.PEN, .ERASER, .BRUSH, .PENCIL, ou .AIRBRUSH,
		# et aussi (même si jsuis pas sûr ce soit pertinent) .UNKNOWN, .MOUSE et
		# .LENS (fuck ces bourgeois avec des tablettes haut de gamme au pire).
		# On pourrait adapter le comportement (couleur/opérateur/aliasing/etc.)
		# à cette information à l'avenir.

		pressure = event.get_axis(Gdk.AxisUse.PRESSURE)
		# print(pressure)
		if pressure is None:
			return None
		return pressure

	############################################################################

	def build_operation(self):
		operation = {
			'tool_id': self.id,
			'rgba': self.main_color,
			'operator': self.operator2,
			'mode': self._selected_mode,
			'line_width': self.tool_width,
			'line_cap': cairo.LineCap.ROUND,
			'line_join': cairo.LineJoin.ROUND,
			'antialias': self._use_antialias,
			'is_preview': True,
			'path': self._manual_path
		}
		return operation

	def do_tool_operation(self, operation):
		if operation['path'] is None or len(operation['path']) < 1:
			return
		cairo_context = self.start_tool_operation(operation)
		cairo_context.set_line_cap(operation['line_cap'])
		cairo_context.set_line_join(operation['line_join'])
		cairo_context.set_source_rgba(*operation['rgba'])

		if operation['mode'] == 'pressure':
			if operation['is_preview']: # Previewing helps performance & debug
				operation['line_width'] = int(operation['line_width'] / 2)
				return self.op_simple(operation, cairo_context)
			self.op_pressure(operation, cairo_context)
		elif operation['mode'] == 'smooth':
			if operation['is_preview']: # Previewing helps performance & debug
				return self.op_simple(operation, cairo_context)
			self.op_smooth(operation, cairo_context)
		elif operation['mode'] == 'macro-w':
			self.op_macro_w(operation, cairo_context)
		else:
			self.op_simple(operation, cairo_context)

	############################################################################

	def op_macro_w(self, operation, cairo_context):
		"""Trying to study whatever tf is the rendering issue #337"""
		cairo_context.set_antialias(cairo.Antialias.DEFAULT)
		cairo_context.set_operator(cairo.Operator.HSL_HUE)
		cairo_context.set_line_width(operation['line_width'])
		cairo_context.new_path()
		cairo_context.line_to(50, 70)
		cairo_context.line_to(70, 50)
		cairo_context.line_to(90, 60)
		cairo_context.line_to(110, 85)
		cairo_context.line_to(130, 120)
		cairo_context.line_to(150, 110)
		cairo_context.line_to(170, 80)
		cairo_context.line_to(190, 50)
		cairo_context.line_to(210, 75)
		cairo_context.line_to(230, 90)
		cairo_context.line_to(250, 110)
		cairo_context.line_to(270, 70)
		cairo_context.line_to(290, 50)
		cairo_context.line_to(310, 75)
		cairo_context.line_to(330, 90)
		cairo_context.stroke()

	############################################################################

	def op_simple(self, operation, cairo_context):
		cairo_context.set_operator(operation['operator'])
		cairo_context.set_line_width(operation['line_width'])
		cairo_context.new_path()
		for pt in operation['path']:
			cairo_context.line_to(pt['x'], pt['y'])
		cairo_context.stroke()

	def op_smooth(self, operation, cairo_context):
		cairo_context.set_operator(cairo.Operator.OVER)
		cairo_context.set_line_width(operation['line_width'])

		# Build a cairo path from the raw data
		cairo_context.new_path()
		for pt in operation['path']:
			cairo_context.line_to(pt['x'], pt['y'])
		path = cairo_context.copy_path()

		# Smooth it
		cairo_context.new_path()
		utilities_smooth_path(cairo_context, path)

		# Draw it
		cairo_context.stroke()

	############################################################################
################################################################################




# tool_scale.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GdkPixbuf, GLib
from .abstract_transform_tool import AbstractCanvasTool
from .optionsbar_scale import OptionsBarScale
from .utilities_overlay import utilities_show_handles_on_context

class ToolScale(AbstractCanvasTool):
	__gtype_name__ = 'ToolScale'

	def __init__(self, window):
		super().__init__('scale', _("Scale"), 'tool-scale-symbolic', window)
		self.cursor_name = 'not-allowed'

		# depends on both the option AND the click coordinates
		self._preserve_ratio = True

		# the user inverted the ratio preservation behavior with a modifier key
		self._ratio_is_inverted = False

		# safety lock to reload asynchronously
		self._reload_is_locked = False

		# safety lock to set values in the spinbuttons
		self._spinbtns_disabled = True

		self._x = 0
		self._y = 0
		self._x2 = 0
		self._y2 = 0
		self.x_press = self.x_motion = None
		self.y_press = self.y_motion = None
		self.add_tool_action_enum('scale-proportions', 'corners')
		self.add_tool_action_boolean('scale-ratio-spinbtns', False)
		self.add_tool_action_enum('scale-unit', 'pixels')

	def try_build_pane(self):
		self.pane_id = 'scale'
		self.window.options_manager.try_add_bottom_pane(self.pane_id, self)

	def build_bottom_pane(self):
		bar = OptionsBarScale(self)

		self._width_btn = bar.width_btn
		self._height_btn = bar.height_btn
		self._width_btn.connect('value-changed', self.on_spinbtn_changed)
		self._height_btn.connect('value-changed', self.on_spinbtn_changed)

		self._w100_btn = bar.width_100_btn
		self._h100_btn = bar.height_100_btn
		self._w100_btn.connect('value-changed', self.on_spinbtn100_changed)
		self._h100_btn.connect('value-changed', self.on_spinbtn100_changed)

		return bar

	############################################################################

	def get_options_label(self):
		return _("Scaling options")

	def get_editing_tips(self):
		# The current method is called when options change, so i use it to:

		# 1) update the visibility of the spinbuttons depending on which ones
		# the user wants to use.
		self._update_to_wanted_unit()

		# 2) update the bullshit around ratio preservation actions
		ratio_option = self.get_option_value('scale-proportions')
		if not ratio_option == 'corners':
			action = self.window.lookup_action('scale-ratio-spinbtns')
			action.set_state(GLib.Variant.new_boolean(ratio_option == 'always'))
		self.set_action_sensitivity('scale-ratio-spinbtns', ratio_option == 'corners')

		if self.apply_to_selection:
			label_action = _("Scaling the selection")
			label_confirm = None
		else:
			label_action = _("Scaling the canvas")
			label_confirm = self.label + " - " + \
			                         _("Don't forget to confirm the operation!")

		if self.get_image().get_mouse_is_pressed():
			label_modifier_shift = None
		else:
			label_modifier_shift = _("Press <Shift> to quickly toggle the " + \
			                                        "'lock proportions' option")

		# there is intentionally no `label_direction` because i expect the users
		# to understand by themselves how it works.

		full_list = [label_action, label_confirm, label_modifier_shift]
		return list(filter(None, full_list))

	def on_tool_selected(self, *args):
		super().on_tool_selected()
		self._x = 0
		self._y = 0
		self.x_press = self.x_motion = None
		self.y_press = self.y_motion = None
		self._w100_btn.set_value(100)
		self._h100_btn.set_value(100)

		width, height = self._get_original_size()
		self.set_preserve_ratio()
		self._ratio = width / height
		self._spinbtns_disabled = False
		self._width_btn.set_value(width)
		self._height_btn.set_value(height)
		self.build_and_do_op() # Ensure a correct preview

	def _get_original_size(self):
		if self.apply_to_selection:
			original_pixbuf = self.get_selection_pixbuf()
		else:
			original_pixbuf = self.get_main_pixbuf()
		return original_pixbuf.get_width(), original_pixbuf.get_height()

	############################################################################

	def _update_to_wanted_unit(self):
		want_pixels = self.get_option_value('scale-unit') == 'pixels'
		if want_pixels:
			self._width_btn.set_visible(True)
			self._height_btn.set_visible(True)
			self._w100_btn.set_visible(False)
			self._h100_btn.set_visible(False)
		else:
			self._width_btn.set_visible(False)
			self._height_btn.set_visible(False)
			self._w100_btn.set_visible(True)
			self._h100_btn.set_visible(True)

	def set_preserve_ratio(self, for_spinbtns=False):
		"""Set whether or not `self._preserve_ratio` should be true. If it is,
		AND that wasn't already the case before, this method sets the value of
		`self._ratio` too.
		The parameter is `True` if the request to set the ratio is triggered by
		a direct interaction with a spinbutton: the method will not look at the
		same options to decide what to do."""
		former_setting = self._preserve_ratio

		if for_spinbtns:
			self._preserve_ratio = self.get_option_value('scale-ratio-spinbtns')
		else:
			setting = self.get_option_value('scale-proportions')
			if setting == 'corners':
				self._preserve_ratio = len(self._directions) != 1
			else:
				self._preserve_ratio = setting == 'always'

		if self._ratio_is_inverted:
			self._preserve_ratio = not self._preserve_ratio

		if self._preserve_ratio == former_setting:
			return
		if self._preserve_ratio:
			self._ratio = self._get_width() / self._get_height()

	def _try_scale_dimensions(self):
		if self._reload_is_locked:
			return
		self._reload_is_locked = True
		GLib.timeout_add(100, self._unlock_reload, {})
		self._async_try_scale_dimensions()

	def _unlock_reload(self, content_params):
		self._reload_is_locked = False
		self._async_try_scale_dimensions()

	def _async_try_scale_dimensions(self, data_dict={}):
		"""When the value in a spinbutton changes, adjust the values in the
		spinbuttons if necessary, and build-and-do the corresponding tool
		operation."""
		self._spinbtns_disabled = True

		if self._preserve_ratio:
			temp_pixbuf = self.get_image().temp_pixbuf
			existing_width = temp_pixbuf.get_width()
			existing_height = temp_pixbuf.get_height()

			new_width = self._get_width()
			new_height = self._get_height()

			if existing_width != new_width:
				new_height = new_width / self._ratio
				self._height_btn.set_value(new_height)
			if existing_height != new_height:
				new_width = new_height * self._ratio
				self._width_btn.set_value(new_width)
		# else:
			# If the ratio isn't locked, the dimension should be applied without
			# any change. Calculations to adjust the size are only useful when
			# trying to preserve the image proportions.

		# Update the "percentage" spinbuttons
		original_width , original_height = self._get_original_size()
		new_w100 = 100 * (self._get_width() / original_width)
		new_h100 = 100 * (self._get_height() / original_height)
		self._w100_btn.set_value(new_w100)
		self._h100_btn.set_value(new_h100)

		self._spinbtns_disabled = False
		self.build_and_do_op()

	############################################################################
	# Spinbuttons management ###################################################

	def on_spinbtn_changed(self, *args):
		if self._spinbtns_disabled:
			return
		if self.x_press is None:
			# This means the user interacts with the spinbtn directly, instead
			# of the surface.
			self.set_preserve_ratio(True)
		self._try_scale_dimensions()

	def on_spinbtn100_changed(self, *args):
		if self._spinbtns_disabled:
			return
		# This callback can only pass its guard clause when triggered from the
		# spinbutton directly: i don't have to check the value of self.x_press
		self.set_preserve_ratio(True)

		original_width, original_height = self._get_original_size()
		new_width = original_width * (self._w100_btn.get_value() / 100)
		new_height = original_height * (self._h100_btn.get_value() / 100)
		delta_x = self._get_width() - new_width
		delta_y = self._get_height() - new_height
		self._apply_deltas_to_spinbtns(new_width, new_height, delta_x, delta_y)

	def _apply_deltas_to_spinbtns(self, new_width, new_height, dx=0, dy=0):
		if self._preserve_ratio:
			if abs(dy) > abs(dx):
				self._height_btn.set_value(new_height)
			else:
				self._width_btn.set_value(new_width)
		else:
			self._height_btn.set_value(new_height)
			self._width_btn.set_value(new_width)

	def _get_width(self):
		return self._width_btn.get_value_as_int()

	def _get_height(self):
		return self._height_btn.get_value_as_int()

	############################################################################

	def on_unclicked_motion_on_area(self, event, surface):
		self.set_directional_cursor(event.x, event.y)

	def on_press_on_area(self, event, surface, event_x, event_y):
		self.update_modifier_state(event.state)
		if 'SHIFT' in self._modifier_keys:
			# The value will be restored later in `on_release_on_area`
			self._ratio_is_inverted = True

		self.x_press = self.x_motion = event_x
		self.y_press = self.y_motion = event_y
		self._x2 = self._x + self._get_width()
		self._y2 = self._y + self._get_height()
		self.set_preserve_ratio()

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		if self._directions == '':
			return
		delta_x = event_x - self.x_motion
		delta_y = event_y - self.y_motion
		self.x_motion = event_x
		self.y_motion = event_y

		height = self._get_height()
		width = self._get_width()
		if 'n' in self._directions:
			height -= delta_y
			self._y = self._y + delta_y
		if 's' in self._directions:
			height += delta_y
		if 'w' in self._directions:
			width -= delta_x
			self._x = self._x + delta_x
		if 'e' in self._directions:
			width += delta_x

		if self.apply_to_selection and self._preserve_ratio:
			if 'w' in self._directions:
				self._x = self._x2 - width
			if 'n' in self._directions:
				self._y = self._y2 - height

		self._apply_deltas_to_spinbtns(width, height, delta_x, delta_y)

	def on_release_on_area(self, event, surface, event_x, event_y):
		self.on_motion_on_area(event, surface, event_x, event_y)
		self._ratio_is_inverted = False
		self.build_and_do_op() # technically already done
		self._scroll_to_end(event_x - self.x_press, event_y - self.y_press)

		# Reset those to tell apart a spinbtn signals' possible origins
		self.x_press = self.x_motion = None
		self.y_press = self.y_motion = None

	############################################################################

	def on_draw_above(self, area, cairo_context):
		if self.apply_to_selection:
			x1 = int(self._x)
			y1 = int(self._y)
		else:
			x1 = 0
			y1 = 0
		x2 = x1 + self._get_width()
		y2 = y1 + self._get_height()
		x1, x2, y1, y2 = self.get_image().get_corrected_coords(x1, x2, y1, y2, \
		                                         self.apply_to_selection, False)
		self._draw_temp_pixbuf(cairo_context, x1, y1)
		thickness = self.get_overlay_thickness()
		utilities_show_handles_on_context(cairo_context, x1, x2, y1, y2, thickness)

	############################################################################

	def build_operation(self):
		operation = {
			'tool_id': self.id,
			'is_selection': self.apply_to_selection,
			'is_preview': True,
			'local_dx': int(self._x),
			'local_dy': int(self._y),
			'width': self._get_width(),
			'height': self._get_height()
		}
		return operation

	def do_tool_operation(self, operation):
		self.start_tool_operation(operation)
		if operation['is_selection']:
			source_pixbuf = self.get_selection_pixbuf()
		else:
			source_pixbuf = self.get_main_pixbuf()
		self.get_image().set_temp_pixbuf(source_pixbuf.scale_simple( \
		                                 operation['width'], \
		                                 operation['height'], \
		                                 GdkPixbuf.InterpType.TILES))
		self.common_end_operation(operation)

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

class AbstractFilter():
	__gtype_name__ = 'AbstractFilter'

	def __init__(self, filter_id, filters_tool, *args):
		self._id = filter_id
		self._tool = filters_tool

	def get_preferred_minimum_width(self):
		return 0

	def set_filter_compact(self, is_active, is_compact):
		pass

	def set_attributes_values(self):
		pass

	def build_filter_op(self):
		return {}

	def do_filter_operation(self, source_pixbuf, operation):
		pass

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

from gi.repository import Gtk, Gio

################################################################################

def utilities_add_filechooser_filters(dialog):
	"""Add file filters for images to file chooser dialogs."""
	allPictures = Gtk.FileFilter()
	allPictures.set_name(_("All pictures"))
	allPictures.add_mime_type('image/png')
	allPictures.add_mime_type('image/jpeg')
	allPictures.add_mime_type('image/bmp')

	pngPictures = Gtk.FileFilter()
	pngPictures.set_name(_("PNG images"))
	pngPictures.add_mime_type('image/png')

	jpegPictures = Gtk.FileFilter()
	jpegPictures.set_name(_("JPEG images"))
	jpegPictures.add_mime_type('image/jpeg')

	bmpPictures = Gtk.FileFilter()
	bmpPictures.set_name(_("BMP images"))
	bmpPictures.add_mime_type('image/bmp')

	dialog.add_filter(allPictures)
	dialog.add_filter(pngPictures)
	dialog.add_filter(jpegPictures)
	dialog.add_filter(bmpPictures)

################################################################################

def utilities_gfile_is_image(gfile, error_msg=""):
	try:
		infos = gfile.query_info('standard::*', Gio.FileQueryInfoFlags.NONE, None)
		if 'image/' in infos.get_content_type():
			# The exact file format of the image isn't validated here because i
			# can't assume what gdkpixbuf is able to read (it's modular, and it
			# evolves). An InvalidFileFormatException will be raised by DrImage
			# if the file can't be loaded.
			return True, error_msg
		else:
			error_msg = error_msg + _("%s isn't an image.") % gfile.get_path()
	except Exception as err:
		error_msg = error_msg + err.message
	return False, error_msg

class InvalidFileFormatException(Exception):
	def __init__(self, initial_message, fpath):
		self.message = initial_message
		cpt = 0
		with open(fpath, 'rb') as f:
			riff_bytes = f.read(4)
			size_bytes = f.read(4)
			webp_bytes = f.read(4)
			if riff_bytes == b'RIFF' and webp_bytes == b'WEBP':
				msg = _("Sorry, WEBP images can't be loaded by this app.") + " 😢 "
				if fpath[-5:] != '.webp':
					# Context: an error message, %s is a file path
					msg = msg + _("Despite its name, %s is a WEBP file.") % fpath
				self.message = msg
		super().__init__(self.message)
	# This exception is meant to be raised when a file is detected as an image
	# by Gio (see utility function above) BUT can't be loaded into a GdkPixbuf.
	# It usually means the file is corrupted, or has a deceptive name (for
	# example "xxx.jpeg" despite being a text file), or is just an image format
	# not supported by the GdkPixbuf version installed by the user.

################################################################################




# abstract_tool.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from gi.repository import Gtk, Gdk

class WrongToolIdException(Exception):
	def __init__(self, expected, actual):
		# Context: an error message
		msg = _("Can't start operation: wrong tool id (expected {0}, got {1})")
		super().__init__(msg.format(expected, actual))

################################################################################

class AbstractAbstractTool():
	"""Super-class implemented and extended by all tools."""

	__gtype_name__ = 'AbstractAbstractTool'
	UI_PATH = '/com/github/maoschanz/drawing/tools/ui/'

	def __init__(self, tool_id, label, icon_name, window, **kwargs):
		self.window = window
		# The tool's identity
		self.id = tool_id
		self.menu_id = 0
		self.label = self._mnemo_label = label
		self._mnemo_char = ""
		self._icon_name = icon_name
		# The options it supports
		self.accept_selection = False
		self.use_color = False
		self.use_operator = False
		# The tool's state
		self.cursor_name = 'cell'
		self._ongoing_operation = False
		self._modifier_keys = []
		# Once everything is set, build the UI
		self.try_build_pane()

	############################################################################
	# Utilities managing actions for tool's options ############################

	def add_tool_action_simple(self, action_name, callback):
		"""Convenient wrapper method adding a stateless action to the window. It
		will be named 'action_name' (string) and activating the action will
		trigger the method 'callback'."""
		# XXX allow to set shortcuts here?
		self.window.add_action_simple(action_name, callback)

	def add_tool_action_boolean(self, action_name, default):
		self.window.options_manager.add_option_boolean(action_name, default)

	def add_tool_action_enum(self, action_name, default):
		self.window.options_manager.add_option_enum(action_name, default)

	def load_tool_action_boolean(self, action_name, key_name):
		om = self.window.options_manager
		return om.add_option_from_bool_key(action_name, key_name)

	def load_tool_action_enum(self, action_name, key_name):
		om = self.window.options_manager
		return om.add_option_from_enum_key(action_name, key_name)

	def get_option_value(self, action_name):
		return self.window.options_manager.get_value(action_name)

	def set_action_sensitivity(self, action_name, state):
		self.window.lookup_action(action_name).set_enabled(state)

	def update_actions_state(self):
		self.set_action_sensitivity('main_color', self.use_color)
		self.set_action_sensitivity('secondary_color', self.use_color)
		self.set_action_sensitivity('exchange_color', self.use_color)
		self.set_action_sensitivity('cairo_operator', self.use_operator)

		self.set_action_sensitivity('cancel_transform', False)
		self.set_action_sensitivity('apply_transform', False)

	def get_settings(self):
		return self.window.options_manager._tools_gsettings

	def update_modifier_state(self, event_state):
		modifier_keys = []
		# CONTROL_MASK can't be used, because it already has an effect app-wide.
		if (event_state & Gdk.ModifierType.SHIFT_MASK) == Gdk.ModifierType.SHIFT_MASK:
			modifier_keys.append('SHIFT')
		if (event_state & Gdk.ModifierType.MOD1_MASK) == Gdk.ModifierType.MOD1_MASK:
			modifier_keys.append('ALT')
		self._modifier_keys = modifier_keys

	############################################################################
	# Various utilities ########################################################

	def show_error(self, error_text):
		self.window.reveal_message(error_text)

	def get_tooltip(self, event_x, event_y, motion_behavior):
		return None

	############################################################################
	# Bottom pane and menubar integration ######################################

	def try_build_pane(self):
		pass

	def build_bottom_pane(self):
		return None

	def on_apply_transform_tool_operation(self, *args):
		pass # implemented only by transform tools

	def on_cancel_transform_tool_operation(self, *args):
		pass # implemented only by transform tools

	def get_options_label(self):
		return _("No options")

	def adapt_to_window_size(self, available_width):
		pass

	def add_item_to_menu(self, tools_menu):
		tools_menu.append(self._mnemo_label, 'win.active_tool::' + self.id)

	def get_options_model(self):
		"""Returns a Gio.MenuModel corresponding to the tool's options. It'll be
		shown in the menubar (if any) and in the bottom pane (if the tool's
		bottom pane supports such a feature)."""
		fpath = self.UI_PATH + 'tool-' + self.id + '.ui'
		builder = Gtk.Builder.new_from_resource(fpath)
		return builder.get_object('options-menu')

	def get_options_widget(self):
		"""Returns a Gtk.Widget (normally a box) corresponding to the tool's
		options. It'll be in the bottom pane (if the tool's bottom pane supports
		such a feature) in replacement of the Gio.MenuModel if such a simple
		menu can't provide all the features."""
		return None

	def get_editing_tips(self):
		return [self.label]

	############################################################################
	# Side pane ################################################################

	def set_mnemonics(self, character):
		self._mnemo_label = self.label.replace(character, "_" + character, 1)
		self._mnemo_char = character.upper()

	def build_flowbox_child(self, flowbox):
		"""Build the icon and its label for the sidebar."""
		# The icon
		if self.window.gsettings.get_boolean('big-icons'):
			size = Gtk.IconSize.LARGE_TOOLBAR
		else:
			size = Gtk.IconSize.SMALL_TOOLBAR
		image = Gtk.Image().new_from_icon_name(self._icon_name, size)
		self._label_box = Gtk.Box( \
			orientation=Gtk.Orientation.HORIZONTAL, \
			spacing=8, \
			margin=8 \
		)
		self._label_box.add(image)

		# The readable label
		label_widget = Gtk.Label(use_underline=True, label=self._mnemo_label)
		self._label_box.add(label_widget)
		self._label_box.show_all()

		# The "mini-label" is shown only when pressing <Alt>
		mini_label = Gtk.Label(use_underline=True, label="_" + self._mnemo_char)
		self._label_box.add(mini_label)

		flowbox.add(self._label_box)
		self._fb_child = self._label_box.get_parent()

		# The item's tooltip
		if self._mnemo_char == "":
			self._fb_child.set_tooltip_text(self.label)
		else:
			tooltip = self.label + " (Alt+" + self._mnemo_char + ")"
			self._fb_child.set_tooltip_text(tooltip)

	def select_flowbox_child(self, *args):
		self.window.tools_flowbox.select_child(self._fb_child)

	def is_flowbox_child_selected(self):
		return self._fb_child.is_selected()

	def set_show_label(self, label_visible):
		label_widget = self._label_box.get_children()[1]
		label_widget.set_visible(label_visible)
		icon = self._label_box.get_children()[0]
		if label_visible:
			icon.set_halign(Gtk.Align.START)
		else:
			icon.set_halign(Gtk.Align.CENTER)

	def update_icon_size(self):
		image = self._label_box.get_children()[0]
		if self.window.gsettings.get_boolean('big-icons'):
			size = Gtk.IconSize.LARGE_TOOLBAR
		else:
			size = Gtk.IconSize.SMALL_TOOLBAR
		image.set_from_icon_name(self._icon_name, size)

	def show_only_mnemonics(self, should_show):
		if self._mnemo_char == "":
			return
		label_widget = self._label_box.get_children()[1]
		if label_widget.get_visible():
			return
		image = self._label_box.get_children()[0]
		mini_label = self._label_box.get_children()[2]

		image.set_visible(not should_show)
		mini_label.set_visible(should_show)

	############################################################################
	# Activation or not ########################################################

	def on_tool_selected(self):
		pass

	def on_tool_unselected(self):
		pass

	def cancel_ongoing_operation(self):
		"""Reset the current tool when 'undo' is pressed while an operation has
		been started but not applied."""
		self.on_tool_unselected()
		self.give_back_control(self.accept_selection) # XXX pas sûr
		self.on_tool_selected()
		self.restore_pixbuf()
		self.non_destructive_show_modif()
		self._ongoing_operation = False

	def give_back_control(self, preserve_selection):
		"""Reset the tool's state well enough to allow for a new tool to start.
		It should include unselecting/applying the selection if the next tool
		doesn't support it."""
		self.restore_pixbuf()
		self.non_destructive_show_modif()

	############################################################################
	# History ##################################################################

	def has_ongoing_operation(self):
		return self._ongoing_operation

	def do_tool_operation(self, operation):
		pass

	def start_tool_operation(self, operation):
		if operation['tool_id'] != self.id:
			raise WrongToolIdException(operation['tool_id'], self.id)
		self.restore_pixbuf()
		self._ongoing_operation = True

	def apply_operation(self, operation):
		"""Complete method to apply an operation: the operation is applied and
		the image is updated as well as the state of actions."""
		self.simple_apply_operation(operation)
		self.get_image().update_actions_state()
		self.get_image().update_history_sensitivity()

	def simple_apply_operation(self, operation):
		"""Simpler apply_operation, for the 'rebuild from history' method."""
		try:
			self.do_tool_operation(operation)
			self.get_image().add_to_history(operation)
		except Exception as e:
			self.show_error(str(e))
		self._ongoing_operation = False
		self.non_destructive_show_modif() # XXX nécessaire ?

	############################################################################
	# Selection ################################################################

	def get_selection(self):
		return self.get_image().selection

	def selection_is_active(self):
		return self.get_selection().is_active

	def get_selection_pixbuf(self):
		return self.get_selection().get_pixbuf()

	def get_overlay_thickness(self):
		return (1 / self.get_image().zoom_level)

	############################################################################
	# Image management #########################################################

	def get_image(self):
		return self.window.get_active_image()

	def get_surface(self):
		return self.get_image().get_surface()

	def scale_factor(self):
		return self.get_image().SCALE_FACTOR

	def get_context(self):
		return cairo.Context(self.get_surface())

	def get_main_pixbuf(self):
		return self.get_image().main_pixbuf

	def non_destructive_show_modif(self):
		self.get_image().update()

	def restore_pixbuf(self):
		self.get_image().use_stable_pixbuf()

	############################################################################
	# Signals handling #########################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		pass

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		pass

	def on_unclicked_motion_on_area(self, event, surface):
		pass

	def on_release_on_area(self, event, surface, event_x, event_y):
		pass

	def on_draw_above(self, area, cairo_context):
		pass

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

from gi.repository import Gdk, Gio, GdkPixbuf, GLib
# from .abstract_tool import WrongToolIdException

################################################################################

class DrHistoryManager():
	__gtype_name__ = 'DrHistoryManager'

	def __init__(self, image, **kwargs):
		self._image = image

		self._undo_history = []
		self._redo_history = []
		self._is_saved = True
		self._waiting_for_rebuild = False

	def get_saved(self):
		# XXX undoing/redoing doesn't update the title so the "*" isn't visible
		# in all situations around a saving
		return self._is_saved

	def empty_history(self):
		"""Probably useless way to explicitly 'forget' the objects. It doesn't
		really free the memory, but it kinda helps i suppose."""
		for op in self._undo_history:
			self._delete_operation(op)
		for op in self._redo_history:
			self._delete_operation(op)
		self._delete_operation(self.initial_operation)

	def _delete_operation(self, op):
		for key in op:
			op[key] = None
		op = {}
		op = None

	############################################################################
	# Controls accessed by DrImage #############################################

	def try_undo(self):
		if self._operation_is_ongoing():
			self._image.active_tool().cancel_ongoing_operation()
			return
		if len(self._undo_history) > 0:
			last_op = self._undo_history.pop()
			self._redo_history.append(last_op)
		self._rebuild_from_history_async()
		self._image.update_history_sensitivity()

	def try_redo(self, *args):
		operation = self._redo_history.pop()
		if operation['tool_id'] is None:
			self._undo_history.append(operation)
			self._image.restore_last_state()
		else:
			self._get_tool(operation['tool_id']).apply_operation(operation)

	def can_undo(self):
		# XXX incorrect si ya des states et qu'on redo
		return (len(self._undo_history) > 0) or self._operation_is_ongoing()

	def can_redo(self):
		# XXX incorrect si ya des states ?
		return len(self._redo_history) > 0

#	def update_history_actions_labels(self):
#		"""Update the current decoration manager "undo/redo" labels (tooltips,
#		menuitems, etc.) to fit the current content of the history."""
#		# l'appel est commenté aussi
#		undoable_action = self._undo_history[-1:]
#		redoable_action = self._redo_history[-1:]
#		undo_label = None
#		redo_label = None
#		# TODO store/get translatable labels instead of tool_ids (issue #42),
#		# but it'll doesn't work with pixbuf states anyway…
#		if self._operation_is_ongoing():
#			# XXX pointless: the method is called after applying the operation
#			undo_label = self._image.active_tool().tool_id
#		elif len(undoable_action) > 0:
#			undo_label = undoable_action[0]['tool_id']
#		if len(redoable_action) > 0:
#			redo_label = redoable_action[0]['tool_id']
#		self._image.window.update_history_actions_labels(undo_label, redo_label)

	def rewind_history(self):
		"""Put the entire 'undo' history into the 'redo' history, so the image
		can be reset without losing any data."""
		self._redo_history = self._undo_history[::-1] + self._redo_history
		self._undo_history = []
		self._image.update_history_sensitivity()

	############################################################################
	# Serialized operations ####################################################

	def add_operation(self, operation):
		self._image.set_surface_as_stable_pixbuf()
		# print('add_operation_to_history')
		# print(operation['tool_id'])
		# if 'select' in operation['tool_id']:
		# 	print(operation['operation_type'])
		# 	print('-----------------------------------')
		self._is_saved = False
		self._undo_history.append(operation)

	############################################################################
	# Cached pixbufs ###########################################################

	def set_initial_operation(self, rgba_array, pixbuf, width, height):
		r = float(rgba_array[0])
		g = float(rgba_array[1])
		b = float(rgba_array[2])
		a = float(rgba_array[3])
		self.initial_operation = {
			'tool_id': None,
			'pixbuf': pixbuf,
			'rgba': Gdk.RGBA(red=r, green=g, blue=b, alpha=a),
			'width': width, 'height': height
		}

	def add_state(self, pixbuf):
		if pixbuf is None:
			# Context: an error message
			raise Exception(_("Attempt to save an invalid state"))
		self._undo_history.append({
			'tool_id': None,
			'pixbuf': pixbuf,
			'width': pixbuf.get_width(),
			'height': pixbuf.get_height()
		})
		self._is_saved = True

	def has_initial_pixbuf(self):
		return self.initial_operation['pixbuf'] is not None

	def get_last_saved_state(self):
		index = self._get_last_state_index(False)
		if index == -1:
			return self.initial_operation
		else:
			return self._undo_history[index]

	def _get_last_state_index(self, allow_yeeting_states):
		"""Return the index of the last "state" operation (dict whose 'tool_id'
		value is None) in the undo-history. If there is no such operation, the
		returned index is -1 which means the only known state is the
		self.initial_operation attribute."""

		returned_index = -1
		nbPixbufs = 0
		for op in self._undo_history:
			if op['tool_id'] is None:
				last_saved_pixbuf_op = op
				returned_index = self._undo_history.index(op)
				nbPixbufs += 1

		# TODO if there are too many pixbufs in the history, remove a few ones
		# Issue #200, needs more design because saving can change the data

		# print("returned_index : " + str(returned_index))
		return returned_index

	############################################################################
	# Other private methods ####################################################

	def _rebuild_from_history_async(self):
		if not self._waiting_for_rebuild:
			# No need to duplicate calls, it will be rebuilt soon anyway
			self._waiting_for_rebuild = True
			GLib.timeout_add(500, self._rebuild_from_history, {})
		# This introduces an artificial delay of half a second, BUT in the case
		# where an user undoes several operations, i expect between 2 and 4
		# presses on ctrl+z during these 500ms, so there will be between 2 and 4
		# times less recomputation.

	def _rebuild_from_history(self, async_cb_data={}):
		"""Rebuild the image according to the content of the current history.
		This is used as a GSourceFunc so it should return False."""
		if not self._waiting_for_rebuild:
			# It has already been rebuild by an other async call
			return False
		self._waiting_for_rebuild = False

		last_save_index = self._get_last_state_index(True)
		self._image.restore_last_state()
		history = self._undo_history.copy()
		self._undo_history = []
		for op in history:
			if history.index(op) > last_save_index:
				# print("do", op['tool_id'])
				self._get_tool(op['tool_id']).simple_apply_operation(op)
			else:
				# print("skip", op['tool_id'])
				self._undo_history.append(op)
		self._image.update()
		return False

	def _operation_is_ongoing(self):
		return self._image.active_tool().has_ongoing_operation()

	def _get_tool(self, tool_id):
		all_tools = self._image.window.tools
		if tool_id in all_tools:
			return all_tools[tool_id]
		else:
			self._image.window.reveal_action_report(_("Error: no tool '%s'" % tool_id))
			# no raise: this may happen normally if last_save_index is incorrect

	############################################################################
################################################################################





# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo, math
from gi.repository import Gdk
from .abstract_brush import AbstractBrush
from .utilities_paths import utilities_smooth_path

class BrushSimple(AbstractBrush):
	__gtype_name__ = 'BrushSimple'

	def _get_tips(self, use_pressure, brush_direction):
		label = _("Simple brush") + " - "
		if use_pressure:
			label += _("Width depends on the stylus pressure")
		else:
			label += _("Width depends on the mouse speed")
		return [label]

	def draw_preview(self, operation, cairo_context):
		cairo_context.set_line_cap(cairo.LineCap.ROUND)
		cairo_context.set_line_join(cairo.LineJoin.ROUND)
		cairo_context.set_source_rgba(*operation['rgba'])
		super().draw_preview(operation, cairo_context)

	############################################################################

	def do_brush_operation(self, cairo_context, operation):
		"""Brush with dynamic width, where the variation of width is drawn by a
		succession of segments. If pressure is detected, the width is pressure-
		sensitive, otherwise it's speed-sensitive (with a heavy ponderation to
		make it less ugly)."""

		if operation['is_preview']: # Previewing helps performance & debug
			operation['line_width'] = max(1, int(operation['line_width'] * 0.8))
			return self.draw_preview(operation, cairo_context)

		if len(operation['path']) < 3:
			# XXX minimum 3 points to get minimum 2 segments to avoid "list
			# index out of range" errors when running the for loops
			return

		self.operation_on_mask(operation, cairo_context)

	def do_masked_brush_op(self, cairo_context, operation):
		cairo_context.set_line_cap(cairo.LineCap.ROUND)
		cairo_context.set_line_join(cairo.LineJoin.ROUND)

		# Build a raw path with lines between the points
		cairo_context.new_path()
		for pt in operation['path']:
			cairo_context.line_to(pt['x'], pt['y'])
		raw_path = cairo_context.copy_path()

		if operation['smooth']:
			# When the zoom is less than 400% (no great precision required by
			# the user), this "raw" path is smoothed.
			cairo_context.new_path()
			utilities_smooth_path(cairo_context, raw_path)
			smoothed_path = cairo_context.copy_path()
		else:
			smoothed_path = raw_path

		# Build an array with all the widths for each segment
		widths = self._build_widths(operation['path'], operation['line_width'])

		# Run through the path to manually draw each segment with its width
		i = 0
		cairo_context.new_path()
		for segment in smoothed_path:
			i = i + 1
			ok, future_x, future_y = self._future_point(segment)
			if not ok:
				cairo_context.move_to(future_x, future_y)
				continue
			current_x, current_y = cairo_context.get_current_point()
			cairo_context.set_line_width(widths[i - 1])
			self._add_segment(cairo_context, segment)
			cairo_context.stroke()
			cairo_context.move_to(future_x, future_y)

	############################################################################
	# Private methods ##########################################################

	def _build_widths(self, manual_path, base_width):
		"""Build an array of widths from the raw data, either using the value of
		the pressure or based on the estimated speed of the movement."""
		widths = []
		dists = []
		p2 = None
		for pt in manual_path:
			if pt['p'] is None:
				# No data about pressure
				if p2 is not None:
					dists.append(self._get_dist(pt['x'], pt['y'], p2['x'], p2['y']))
			else:
				# There are data about pressure
				if p2 is not None:
					if p2['p'] == 0 or pt['p'] == 0:
						seg_width = 0
					else:
						seg_width = (p2['p'] + pt['p']) / 2
					# A segment whose 2 points have a 50% pressure shall have a
					# width of "100%" of the base_width, so "base * mean * 2"
					widths.append(base_width * seg_width * 2)
			p2 = pt

		# If nothing in widths, it has to be filled from dists
		if len(widths) == 0:
			min_dist = min(dists)
			max_dist = max(dists)
			temp_width = 0
			for dist in dists:
				new_width = 1 + int(base_width / max(1, 0.05 * dist))
				if temp_width == 0:
					temp_width = (new_width + base_width) / 2
				else:
					temp_width = (new_width + temp_width + temp_width) / 3
				width = max(1, int(temp_width))
				widths.append(width)

		return widths

	def _add_segment(self, cairo_context, pts):
		if pts[0] == cairo.PathDataType.CURVE_TO:
			cairo_context.curve_to(pts[1][0], pts[1][1], pts[1][2], pts[1][3], \
			                                             pts[1][4], pts[1][5])
		elif pts[0] == cairo.PathDataType.LINE_TO:
			cairo_context.line_to(pts[1][0], pts[1][1])

	def _future_point(self, pts):
		if pts[0] == cairo.PathDataType.CURVE_TO:
			return True, pts[1][4], pts[1][5]
		elif pts[0] == cairo.PathDataType.LINE_TO:
			return True, pts[1][0], pts[1][1]
		else: # all paths start with a cairo.PathDataType.MOVE_TO
			return False, pts[1][0], pts[1][1]

	def _get_dist(self, x1, y1, x2, y2):
		dist2 = (x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)
		return math.sqrt(dist2)

	############################################################################
################################################################################




# Licensed under GPL3 https://github.com/maoschanz/drawing/blob/master/LICENSE

import cairo
from .abstract_brush import AbstractBrush

class BrushNib(AbstractBrush):
	__gtype_name__ = 'BrushNib'

	def _get_tips(self, use_pressure, brush_direction):
		label_width = _("Calligraphic nib") + " - "
		if use_pressure:
			label_width += _("Width depends on the stylus pressure")
		else:
			label_width += _("Width depends on the line orientation")

		label_direction = _("Calligraphic nib") + " - "
		if brush_direction == 'right':
			label_direction += _("Right-handed nib")
		elif brush_direction == 'vertical':
			label_direction += _("Vertical nib")
		elif brush_direction == 'left':
			label_direction += _("Left-handed nib")
		else: # brush_direction == 'horizontal':
			label_direction += _("Horizontal nib")

		return [label_width, label_direction]

	def draw_preview(self, operation, cairo_context):
		cairo_context.set_line_cap(cairo.LineCap.BUTT)
		cairo_context.set_line_join(cairo.LineJoin.ROUND)
		cairo_context.set_source_rgba(*operation['rgba'])
		super().draw_preview(operation, cairo_context)

	############################################################################

	def do_brush_operation(self, cairo_context, operation):
		"""Straight-shaped brush, like an orientable feather pen. The width is
		pressure-sensitive but otherwise it's NOT speed sensitive."""

		if operation['is_preview']: # Previewing helps performance & debug
			return self.draw_preview(operation, cairo_context)

		self.operation_on_mask(operation, cairo_context)

	def do_masked_brush_op(self, cairo_context, operation):
		# cairo_context.set_line_cap(cairo.LineCap.BUTT)
		# cairo_context.set_line_join(cairo.LineJoin.BEVEL)
		# cairo_context.set_line_width(1)

		line_width = operation['line_width'] / 2
		two_ways_path = []

		if operation['nib_dir'] == 'left':
			feather_def = {'x': 1, 'y': 1}
		elif operation['nib_dir'] == 'horizontal':
			feather_def = {'x': 1, 'y': 0}
		elif operation['nib_dir'] == 'vertical':
			feather_def = {'x': 0, 'y': 1}
		else: # operation['nib_dir'] == 'right':
			feather_def = {'x': 1, 'y': -1}

		dx_base = feather_def['x'] * line_width
		dy_base = feather_def['y'] * line_width
		use_pressure = operation['path'][0]['p'] is not None
		if not use_pressure:
			dx1 = dx_base
			dy1 = dy_base
			dx2 = dx_base
			dy2 = dy_base

		former_point = None
		for current_point in operation['path']:
			if former_point is not None:
				if use_pressure:
					dx1 = dx_base * former_point['p']
					dy1 = dy_base * former_point['p']
					dx2 = dx_base * current_point['p']
					dy2 = dy_base * current_point['p']
				cairo_context.new_path()
				x = current_point['x'] + dx2
				y = current_point['y'] + dy2
				cairo_context.move_to(x, y)
				x = current_point['x'] + dx2 * -1
				y = current_point['y'] + dy2 * -1
				cairo_context.line_to(x, y)
				x = former_point['x'] + dx1 * -1
				y = former_point['y'] + dy1 * -1
				cairo_context.line_to(x, y)
				x = former_point['x'] + dx1
				y = former_point['y'] + dy1
				cairo_context.line_to(x, y)
				# cairo_context.stroke_preserve() # TODO trouver autre chose
				cairo_context.fill()
			former_point = current_point

	############################################################################
################################################################################




# optionsbar_color_popover.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from gi.repository import Gtk
from .utilities_colors import utilities_get_rgba_name

PREFIX = '/com/github/maoschanz/drawing/'

CAIRO_OP_VALUES = {
	'over': cairo.Operator.OVER,
	'clear': cairo.Operator.CLEAR,

	'source': cairo.Operator.SOURCE,
	'difference': cairo.Operator.DIFFERENCE,

	# "Highlight" submenu
	'multiply': cairo.Operator.MULTIPLY,
	'screen': cairo.Operator.SCREEN,

	'hsl-hue': cairo.Operator.HSL_HUE,
	'hsl-saturation': cairo.Operator.HSL_SATURATION,
	'hsl-color': cairo.Operator.HSL_COLOR,
	'hsl-luminosity': cairo.Operator.HSL_LUMINOSITY,
}

CAIRO_OP_LABELS = {
	'over': _("Normal"),
	'clear': _("Erase"),

	'source': _("Raw source color"),
	'difference': _("Difference"),

	'multiply': _("Highlight"),
	# Context: this is equivalent to "Highlight: Light text on dark background"
	# but it has to be FAR SHORTER so it fits in the color chooser
	'screen': _("Highlight (dark)"),

	'hsl-hue': _("Hue only"),
	'hsl-saturation': _("Saturation only"),
	'hsl-color': _("Hue and saturation"),
	'hsl-luminosity': _("Luminosity only"),
}

################################################################################

class OptionsBarClassicColorPopover(Gtk.Popover):
	__gtype_name__ = 'OptionsBarClassicColorPopover'

	def __init__(self, btn, thumbn, is_left_color, options_manager):
		super().__init__()

		suffix = 'optionsbars/classic/optionsbar-color-popover.ui'
		builder = Gtk.Builder.new_from_resource(PREFIX + suffix)

		main_box = builder.get_object('main-box')
		self.add(main_box)
		self._button = btn
		self._button.set_popover(self)
		self._thumbnail_image = thumbn
		self._options_manager = options_manager

		# These attributes are the same in both oppovers, but this may evolve
		self._operator_enum = cairo.Operator.OVER
		self._operator_label = _("Normal")

		########################################################################
		# Box at the top #######################################################

		if is_left_color:
			title_label = _("Main color")
		else:
			title_label = _("Secondary color")
		self._popover_title = builder.get_object('popover-title')
		self._popover_title.set_label(title_label)

		self._operator_box_1 = builder.get_object('operator-box-start')
		self._operator_box_2 = builder.get_object('operator-box-end')

		operators_radio_group = builder.get_object('op-group')
		self._build_all_operators_groups(operators_radio_group)

		########################################################################
		# Color chooser widget #################################################

		self.color_widget = builder.get_object('color-widget')
		initial_rgba = self._options_manager.get_persisted_color(is_left_color)
		self.color_widget.set_rgba(initial_rgba)
		self.color_widget.connect('notify::rgba', self._on_color_changed)
		self.color_widget.connect('notify::show-editor', self._update_nav_box)

		########################################################################
		# Navigation box at the bottom #########################################

		back_btn = builder.get_object('back-btn')
		self.editor_box = builder.get_object('editor-box')
		back_btn.connect('clicked', self._close_color_editor)

		self._update_nav_box()
		self._on_color_changed()

	def _build_all_operators_groups(self, radio_group):
		self._operators_menus = []
		self._operators_hideable_radiobtns = {
			'over': None,
			'erase': None,
		}
		suffix = 'optionsbars/classic/optionsbar-operator-menus.ui'
		builder = Gtk.Builder.new_from_resource(PREFIX + suffix)

		box = self._build_radios(radio_group, ['multiply', 'screen'])
		menu_model = builder.get_object('highlight-operators-menu')
		btn = self._build_op_menu(menu_model, 'tool-highlight-symbolic')
		btn.set_tooltip_text(_("Highlight"))
		box.add(btn)
		self._operator_box_1.add(box)

		box = self._build_radios(radio_group, ['hsl-hue', 'hsl-saturation',
		                                       'hsl-color', 'hsl-luminosity'])
		menu_model = builder.get_object('hsl-operators-menu')
		btn = self._build_op_menu(menu_model, 'display-brightness-symbolic')
		btn.set_tooltip_text(_("HSL modes"))
		box.add(btn)
		self._operator_box_1.add(box)

		box = self._build_radios(radio_group, ['source', 'difference'])
		menu_model = builder.get_object('other-operators-menu')
		btn = self._build_op_menu(menu_model, 'view-more-symbolic')
		btn.set_tooltip_text(_("Other modes"))
		box.add(btn)
		self._operator_box_1.add(box)

	def _build_radios(self, radio_group, ops_list):
		groupbox = Gtk.Box(visible=True)
		groupbox.get_style_context().add_class('linked')
		for operator_id in ops_list:
			button = Gtk.RadioButton(visible=False, draw_indicator=False)
			button.join_group(radio_group)
			button.set_detailed_action_name ('win.cairo_op_mirror::' + operator_id)

			# To avoid buttons bigger than the popover the label is ellipsized
			button_label = CAIRO_OP_LABELS[operator_id]
			if len(button_label) > 16:
				button.set_tooltip_text(button_label)
				button_label = button_label[:14] + "…"
			button.set_label(button_label)

			self._operators_hideable_radiobtns[operator_id] = button
			groupbox.add(button)
		return groupbox

	def _build_op_menu(self, menu_model, icon_name):
		menu_btn = Gtk.MenuButton(visible=True)
		image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
		menu_btn.set_image(image)
		menu_btn.set_menu_model(menu_model)
		self._operators_menus.append(menu_btn.get_popover())
		return menu_btn

	############################################################################

	def set_operators_available(self, tool_use_operator, op_as_string):
		"""Adapts the availability of operators to the tool's property."""
		self._operator_box_1.set_visible(tool_use_operator)
		self._operator_box_2.set_visible(tool_use_operator)
		self._popover_title.set_visible(not tool_use_operator)
		self.adapt_to_operator(op_as_string)

	def _operator_supports_color(self, op_as_string):
		return not op_as_string == 'clear'

	def adapt_to_operator(self, op_as_string):
		self._operator_enum = CAIRO_OP_VALUES[op_as_string]
		self._operator_label = CAIRO_OP_LABELS[op_as_string]
		# print("adapt to operator :", op_as_string)
		supports_colors = self._operator_supports_color(op_as_string)

		# Close all little menus
		for op_submenu in self._operators_menus:
			op_submenu.popdown()

		# Show only the radio button for the active operator
		for operator_id in self._operators_hideable_radiobtns:
			op_radiobtn = self._operators_hideable_radiobtns[operator_id]
			if op_radiobtn is not None:
				op_radiobtn.set_visible(operator_id == op_as_string)

		self.color_widget.set_sensitive(supports_colors)
		self._set_thumbnail_color(op_as_string)

	def _on_color_changed(self, *args):
		"""When the use clicks on a color in the palette"""
		op_as_string = self._options_manager.get_value('cairo_operator')
		self._set_thumbnail_color(op_as_string)

	def _set_thumbnail_color(self, op_as_string):
		"""Sets the icon and the tooltip of the popover's button. The icon can
		be an actual icon, or a rectangle of color. The tooltip shows the active
		operator (if pertinent), and an approximation of the color name (if
		pertinent)."""
		rgba = self.color_widget.get_rgba()
		red = rgba.red
		green = rgba.green
		blue = rgba.blue
		alpha = rgba.alpha

		# Draw the thumbnail of the button
		if op_as_string == 'clear':
			self._thumbnail_image.set_from_icon_name('tool-eraser-symbolic', \
			                                         Gtk.IconSize.SMALL_TOOLBAR)
		else:
			surface = cairo.ImageSurface(cairo.Format.ARGB32, 16, 16)
			cairo_context = cairo.Context(surface)
			cairo_context.set_source_rgba(red, green, blue, alpha)
			cairo_context.paint()
			# it could explicit whether it's normal/source/difference
			self._thumbnail_image.set_from_surface(surface)

		# Set the tooltip of the button
		tooltip_string = utilities_get_rgba_name(rgba)
		if not self._operator_supports_color(op_as_string):
			tooltip_string = self._operator_label
		elif op_as_string == 'over':
			pass # Normal situation, no need to tell the user
		else:
			tooltip_string = self._operator_label + " - " + tooltip_string
		self._button.set_tooltip_text(tooltip_string)

	############################################################################

	def editor_setting_changed(self, show_editor):
		self.color_widget.props.show_editor = show_editor
		self._update_nav_box()

	def _close_color_editor(self, *args):
		self.color_widget.props.show_editor = False

	def open(self, *args):
		self._button.activate()

	def _update_nav_box(self, *args):
		"""Update the visibility of the 'back to the palette' button."""
		self.editor_box.set_visible(self.color_widget.props.show_editor)

	############################################################################
################################################################################




# abstract_transform_tool.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo
from gi.repository import Gtk, Gdk, GdkPixbuf

from .abstract_tool import AbstractAbstractTool
from .utilities_colors import utilities_gdk_rgba_to_normalized_array

class AbstractCanvasTool(AbstractAbstractTool):
	__gtype_name__ = 'AbstractCanvasTool'

	def __init__(self, tool_id, label, icon_name, window, **kwargs):
		super().__init__(tool_id, label, icon_name, window)
		self.menu_id = 1
		self.centered_box = None
		self.needed_width_for_long = 0
		self.accept_selection = True
		self.apply_to_selection = False
		self._directions = ''

		# Gdk.RGBA
		self._expansion_rgba = None

	def on_tool_selected(self, *args):
		super().on_tool_selected()
		self.apply_to_selection = self.selection_is_active()

	def update_actions_state(self, *args):
		# Changing this in on_tool_selected would be overridden by image.py
		super().update_actions_state()

		self.set_action_sensitivity('selection_delete', False)
		self.set_action_sensitivity('selection_cut', False)
		self.set_action_sensitivity('unselect', False)
		self.set_action_sensitivity('select_all', False)

		self.set_action_sensitivity('cancel_transform', True)
		self.set_action_sensitivity('apply_transform', True)

	def give_back_control(self, preserve_selection):
		if not preserve_selection and self.selection_is_active():
			self.on_apply_transform_tool_operation()
			self.window.get_selection_tool().unselect_and_apply()
		super().give_back_control(preserve_selection)

	def _scroll_to_end(self, h_growth, v_growth):
		if h_growth > 0:
			if 'e' in self._directions:
				self.window.action_go_last()
		if v_growth > 0:
			if 's' in self._directions:
				self.window.action_go_bottom()
		self.get_image().fake_scrollbar_update()

	############################################################################

	def build_and_do_op(self):
		operation = self.build_operation()
		self.do_tool_operation(operation)

	def temp_preview(self, is_selection, local_dx, local_dy):
		"""Part of the previewing methods shared by all transform tools."""
		pixbuf = self.get_image().temp_pixbuf
		cairo_context = self.get_context()
		if is_selection:
			cairo_context.set_source_surface(self.get_surface(), 0, 0)
			cairo_context.paint()
			x = self.get_selection().selection_x + local_dx
			y = self.get_selection().selection_y + local_dy
			Gdk.cairo_set_source_pixbuf(cairo_context, pixbuf, x, y)
			cairo_context.paint()
		else:
			cairo_context.set_operator(cairo.Operator.SOURCE)
			Gdk.cairo_set_source_pixbuf(cairo_context, pixbuf, 0, 0)
			cairo_context.paint()
			cairo_context.set_operator(cairo.Operator.OVER)
		self.get_image().update()

	############################################################################

	def on_apply_transform_tool_operation(self, *args):
		self.restore_pixbuf()
		operation = self.build_operation()
		self.apply_operation(operation)
		self.on_cancel_transform_tool_operation()

	def on_cancel_transform_tool_operation(self, *args):
		if self.apply_to_selection:
			self.window.force_selection()
		else:
			self.window.back_to_previous()

	def apply_operation(self, operation):
		operation['is_preview'] = False
		super().apply_operation(operation)
		self.get_image().reset_temp()

	def common_end_operation(self, op):
		if op['is_preview']:
			self.temp_preview(op['is_selection'], op['local_dx'], op['local_dy'])
		else:
			self.apply_temp(op['is_selection'], op['local_dx'], op['local_dy'])

	def apply_temp(self, operation_is_selection, ldx, ldy):
		new_pixbuf = self.get_image().temp_pixbuf.copy()
		if operation_is_selection:
			self.get_selection().update_from_transform_tool(new_pixbuf, ldx, ldy)
		else:
			self.get_image().set_main_pixbuf(new_pixbuf)
			self.get_image().use_stable_pixbuf()

	############################################################################
	# Direction of the cursor depending on its position ########################

	def _set_directions(self, event_x, event_y, n_sizes):
		"""Set the directions of the user's future operation for tools such as
		`scale` or `crop`.
		It returns a boolean telling whether or not the directions changed since
		the last call to this method."""
		w_left = n_sizes['wl']
		w_right = n_sizes['wr']
		h_top = n_sizes['ht']
		h_bottom = n_sizes['hb']
		# print("set_directions", w_left, w_right, h_top, h_bottom)
		directions = ''
		if event_y < h_top:
			directions += 'n'
		elif event_y > h_bottom:
			directions += 's'
		if event_x < w_left:
			directions += 'w'
		elif event_x > w_right:
			directions += 'e'

		if self._directions == directions:
			return False
		else:
			self._directions = directions
			return True

	def _set_cursor_name(self):
		"""Set the cursor name, depending on the previously set directions."""
		if self._directions == '':
			self.cursor_name = 'not-allowed'
		else:
			self.cursor_name = self._directions + '-resize'

	def set_directional_cursor(self, event_x, event_y):
		"""Set the accurate cursor depending on the position of the pointer on
		the canvas."""
		n_sizes = self.get_image().get_nineths_sizes(self.apply_to_selection, \
		                                                       self._x, self._y)
		# if we're transforming the selection from its top and/or left, coords
		# to decide the direction depend on local deltas (self._x and self._y)
		if self._set_directions(event_x, event_y, n_sizes):
			self._set_cursor_name()
			self.window.set_cursor(True)

	############################################################################

	def on_draw_above(self, area, cairo_context):
		pass

	def _draw_temp_pixbuf(self, cairo_context, x, y):
		pixbuf = self.get_image().temp_pixbuf
		Gdk.cairo_set_source_pixbuf(cairo_context, pixbuf, x, y)
		if self.get_image().is_zoomed_surface_sharp():
			cairo_context.get_source().set_filter(cairo.FILTER_NEAREST)
		cairo_context.paint()

	def get_deformed_surface(self, source_surface, coefs, prefill=False):
		"""Use cairo.Matrix to apply a transformation to `source_surface` using
		the coefficients in `coefs` and return a new surface with the result."""
		p_xx, p_yx, p_xy, p_yy, p_x0, p_y0 = coefs

		source_w = source_surface.get_width()
		source_h = source_surface.get_height()
		w = p_xx * source_w + p_xy * source_h + p_x0 * 2
		h = p_yx * source_w + p_yy * source_h + p_y0 * 2

		new_surface = cairo.ImageSurface(cairo.Format.ARGB32, int(w), int(h))
		cairo_context = cairo.Context(new_surface)

		if prefill:
			# TODO this exists only for the "skew" tool, and it shouldn't be
			# here nor do this.
			# TODO it should paint *around* the deformed surface after it's
			# returned by the current method
			color_array = utilities_gdk_rgba_to_normalized_array(self._expansion_rgba)
			cairo_context.set_source_rgba(*color_array)
			cairo_context.paint()

		# m = cairo.Matrix(xx=1.0, yx=0.0, xy=0.0, yy=1.0, x0=0.0, y0=0.0)
		m = cairo.Matrix(xx=p_xx, yx=p_yx, xy=p_xy, yy=p_yy, x0=p_x0, y0=p_y0)
		try:
			cairo_context.transform(m)
		except:
			self.show_error(_("Error: invalid values"))
			return source_surface
		cairo_context.set_source_surface(source_surface, 0, 0)
		cairo_context.paint()
		return new_surface

	############################################################################
	# Options ##################################################################

	def _update_expansion_rgba(self, event_btn=1):
		"""When the canvas grows, the color of the new pixels is parametrable"""
		color_type = self.get_option_value('crop-expand')
		self._force_expansion_rgba(color_type, event_btn)

	def _force_expansion_rgba(self, color_type, event_btn=1):
		if color_type == 'initial':
			exp_rgba = self.get_image().get_initial_rgba()
		elif color_type == 'secondary' and event_btn == 1:
			exp_rgba = self.window.options_manager.get_right_color()
		elif color_type == 'secondary' and event_btn == 3:
			exp_rgba = self.window.options_manager.get_left_color()
		else: # color_type == 'alpha':
			exp_rgba = Gdk.RGBA(red=1.0, green=1.0, blue=1.0, alpha=0.0)
		self._expansion_rgba = exp_rgba

	############################################################################
################################################################################




# tool_eraser.py
#
# Copyright 2018-2023 Romain F. T.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cairo, random
from gi.repository import Gdk, GdkPixbuf
from .tool_pencil import ToolPencil
from .abstract_classic_tool import AbstractClassicTool
from .utilities_blur import utilities_blur_surface, BlurType, BlurDirection

class ToolEraser(ToolPencil):
	__gtype_name__ = 'ToolEraser'

	def __init__(self, window, **kwargs):
		super().__init__(window)
		# Context: this is the name of a tool, in the meaning "a rubber eraser"
		AbstractClassicTool.__init__(self, 'eraser', _("Eraser"), \
		                                         'tool-eraser-symbolic', window)
		self.use_operator = False
		self._fallback_operator = 'clear'
		self.load_tool_action_enum('selection-color', 'last-delete-replace')
		self.add_tool_action_enum('eraser-type', 'mosaic')
		self.add_tool_action_enum('eraser-shape', 'pencil')
		self._rgba = [0.0, 0.0, 0.0, 0.0]

	def get_editing_tips(self):
		self._rgba_type = self.get_option_value('selection-color')
		self._eraser_shape = self.get_option_value('eraser-shape')
		self._apply_shape_constraints()

		label_options = self.label
		if self._eraser_shape == 'pencil':
			label_options += " - " + _("Pencil")
		else:
			label_options += " - " + _("Rectangle")
		if self._eraser_type == 'solid':
			label_options += " - " + {
				'alpha': _("Transparency"),
				'initial': _("Default color"),
				'secondary': _("Secondary color")
			}[self._rgba_type]
		else:
			label_options += " - " + {
				'blur': _("Blur"),
				'shuffle': _("Shuffle pixels"),
				'mixed': _("Shuffle and blur"),
				'mosaic': _("Mosaic")
			}[self._eraser_type]

		if self.get_image().get_mouse_is_pressed():
			label_modifier_shift = None
		else:
			label_modifier_shift = self.label + " - "
			if self._eraser_shape == 'pencil':
				label_modifier_shift += _("Press <Shift> to erase a rectangle area instead")
			else:
				label_modifier_shift += _("Press <Shift> to erase a path instead")

		full_list = [label_options, label_modifier_shift]
		return list(filter(None, full_list))

	def get_options_label(self):
		return _("Eraser options")

	def _apply_shape_constraints(self):
		self._eraser_type = self.get_option_value('eraser-type')

		can_blur = self._eraser_shape != 'pencil'
		self.set_action_sensitivity('eraser-type', can_blur)
		if not can_blur:
			self._eraser_type = 'solid'

		if 'solid' == self._eraser_type and 'secondary' == self._rgba_type:
			self._fallback_operator = 'source'
		else:
			self._fallback_operator = 'clear'
			# TODO en pratique non il y a des cas où on est plutôt en train de
			# flouter, il faudrait un elif, et un autre système pour l'opérateur
			# en fallback qui afficherait l'icône avec les gouttes.
			# En fait on devrait yeet le délire du `_fallback_operator` ?
		self.window.options_manager.update_pane(self)

	############################################################################

	def on_press_on_area(self, event, surface, event_x, event_y):
		super().on_press_on_area(event, surface, event_x, event_y)
		# TODO the eraser grew up to be too different from the pencil, it
		# shouldn't extend it

		if self._rgba_type == 'alpha':
			self._rgba = [0.0, 0.0, 0.0, 0.0]
		elif self._rgba_type == 'initial':
			clr = self.get_image().get_initial_rgba()
			self._rgba = [clr.red, clr.green, clr.blue, clr.alpha]
		elif self._rgba_type == 'secondary':
			self._rgba = self.secondary_color

		self.update_modifier_state(event.state)
		if 'SHIFT' in self._modifier_keys:
			if self._eraser_shape == 'rectangle':
				self._eraser_shape = 'pencil'
			else:
				self._eraser_shape = 'rectangle'
			self._apply_shape_constraints()

	def on_motion_on_area(self, event, surface, event_x, event_y, render=True):
		if self._eraser_shape == 'rectangle':
			self._draw_rectangle(event_x, event_y)
		else:
			self._add_point(event_x, event_y)

		if not render:
			return
		operation = self.build_operation(True)
		self.do_tool_operation(operation)

	def on_release_on_area(self, event, surface, event_x, event_y):
		if self._eraser_shape == 'rectangle':
			self._draw_rectangle(event_x, event_y)
		else:
			self._add_point(event_x, event_y)
		operation = self.build_operation(False)
		self.apply_operation(operation)
		self._reset_temp_points()

	def _draw_rectangle(self, event_x, event_y):
		cairo_context = self.get_context()
		cairo_context.move_to(self.x_press, self.y_press)
		cairo_context.line_to(self.x_press, event_y)
		cairo_context.line_to(event_x, event_y)
		cairo_context.line_to(event_x, self.y_press)
		cairo_context.close_path()
		self._path = cairo_context.copy_path()

	def _reset_temp_points(self):
		self._path = None
		self.x_press = -1.0
		self.y_press = -1.0

	############################################################################

	def build_operation(self, is_preview):
		if is_preview:
			eraser_type = 'solid'
		else:
			eraser_type = self._eraser_type
		operation = {
			'tool_id': self.id,
			'is_preview': is_preview,
			'line_width': self.tool_width,
			'line_cap': self._cap_id,
			'line_join': self._join_id,
			'replacement': self._rgba,
			'censor-type': eraser_type,
			'censor-shape': self._eraser_shape,
			'antialias': self._use_antialias,
			'path': self._path
		}
		return operation

	def do_tool_operation(self, operation):
		if operation['path'] is None:
			return
		cairo_context = self.start_tool_operation(operation)
		cairo_context.set_operator(cairo.Operator.SOURCE)
		censor_type = operation['censor-type']

		if censor_type == 'solid':
			cairo_context.set_source_rgba(*operation['replacement'])
			if operation['censor-shape'] == 'rectangle':
				cairo_context.append_path(operation['path'])
				cairo_context.fill()
			else:
				cairo_context.set_line_cap(cairo.LineCap.ROUND)
				cairo_context.set_line_join(cairo.LineJoin.ROUND)
				cairo_context.set_line_width(operation['line_width'])
				cairo_context.append_path(operation['path'])
				cairo_context.stroke()
			return
		else:
			cairo_context.append_path(operation['path'])

		[r0, r1, r2, r3] = cairo_context.path_extents()
		[r0, r1, r2, r3] = [int(r0), int(r1), int(r2), int(r3)]
		width = r2 - r0
		height = r3 - r1
		surface = cairo.ImageSurface(cairo.Format.ARGB32, width, height)
		ccontext2 = cairo.Context(surface)
		ccontext2.set_source_surface(self.get_surface(), -1 * r0, -1 * r1)
		ccontext2.paint()
		scale = self.scale_factor()
		surface.set_device_scale(scale, scale)

		b_rad = min(15, int(min(width, height) / 4))
		b_dir = BlurDirection.BOTH
		shuffle_intensity = int((width * height) / 2)
		if censor_type == 'mosaic':
			bs = utilities_blur_surface(surface, b_rad, BlurType.TILES, b_dir)
		elif censor_type == 'blur':
			bs = utilities_blur_surface(surface, b_rad, BlurType.PX_BOX, b_dir)
		elif censor_type == 'shuffle':
			bs = self._shuffle_pixels(surface, shuffle_intensity)
		elif censor_type == 'mixed':
			bs = self._shuffle_pixels(surface, shuffle_intensity / 2)
			bs = utilities_blur_surface(bs, b_rad, BlurType.CAIRO_REPAINTS, b_dir)

		cairo_context.clip()
		# XXX this ^ doesn't work with the 'pencil' shape, which forces me to
		# disable the 'eraser-type' option in this case
		cairo_context.set_source_surface(bs, r0, r1)
		cairo_context.paint()

	############################################################################

	def _shuffle_pixels(self, surface, iterations):
		w = surface.get_width()
		h = surface.get_height()
		channels = 4 # ARGB
		if w <= 1 or h <= 1:
			return surface

		pixels = surface.get_data()

		random.seed(1)
		while iterations > 0:
			iterations = iterations - 1
			self._shuffle_one_iteration(w, h, channels, pixels)
		return surface

	def _shuffle_one_iteration(self, w, h, channels, pixels):
		pix1_x = random.randint(0, w - 1)
		pix1_y = random.randint(0, h - 1)

		# Get data for a first pixel
		cur_pixel = (pix1_y * w + pix1_x) * channels
		a1 = pixels[cur_pixel + 0]
		r1 = pixels[cur_pixel + 1]
		g1 = pixels[cur_pixel + 2]
		b1 = pixels[cur_pixel + 3]

		pix2_x = random.randint(0, w - 1)
		pix2_y = random.randint(0, h - 1)

		# Get data for a second pixel
		cur_pixel = (pix2_y * w + pix2_x) * channels
		a2 = pixels[cur_pixel + 0]
		r2 = pixels[cur_pixel + 1]
		g2 = pixels[cur_pixel + 2]
		b2 = pixels[cur_pixel + 3]

		# Data of the 1st pixel is written in the 2nd one
		pixels[cur_pixel + 0] = a1
		pixels[cur_pixel + 1] = r1
		pixels[cur_pixel + 2] = g1
		pixels[cur_pixel + 3] = b1

		# Data of the 2nd pixel is written in the 1st one
		cur_pixel = (pix1_y * w + pix1_x) * channels
		pixels[cur_pixel + 0] = a2
		pixels[cur_pixel + 1] = r2
		pixels[cur_pixel + 2] = g2
		pixels[cur_pixel + 3] = b2

	############################################################################
################################################################################

